@@ -22,8 +22,7 @@ static int rbnode_replace(rbtree_t *rbtree, rbnode_t *existing, rbnode_t *replac
 	}									\
 } while (0)
 
-__weak
-u64 rb_create_internal(void)
+u64 rb_create_internal(enum rbtree_alloc alloc, enum rbtree_insert_mode insert)
 {
 	rbtree_t *rbtree;
 
@@ -32,6 +31,8 @@ u64 rb_create_internal(void)
 		return (u64)NULL;
 
 	rbtree->root = NULL;
+	rbtree->alloc = alloc;
+	rbtree->insert = insert;
 
 	return (u64)rbtree;
 }
@@ -163,12 +164,13 @@ int rb_find(rbtree_t __arg_arena *rbtree, u64 key, u64 *value)
 	return 0;
 }
 
-__weak
-u64 rb_node_alloc_internal(rbtree_t __arg_arena *rbtree, u64 key, u64 value)
+static inline rbnode_t *rb_node_alloc_common(rbtree_t __arg_arena *rbtree, u64 key, u64 value)
 {
 	rbnode_t *rbnode;
 	volatile rbnode_t *node;
 
+	/* We can't allocate an node for an rbtree that does the allocations itself. */
+
 	do  {
 		rbnode = rbtree->freelist;
 		if (!rbnode)
@@ -179,7 +181,7 @@ u64 rb_node_alloc_internal(rbtree_t __arg_arena *rbtree, u64 key, u64 value)
 	if (!rbnode)
 		rbnode = (rbnode_t *)scx_static_alloc(sizeof(*rbnode), 1);
 	if (!rbnode)
-		return (u64)NULL;
+		return NULL;
 
 	/*
 	 * XXXETSAL:  Use a second volatile variable because the verifier demotes
@@ -196,10 +198,19 @@ u64 rb_node_alloc_internal(rbtree_t __arg_arena *rbtree, u64 key, u64 value)
 	node->value = value;
 	node->is_red = true;
 
-	return (u64)rbnode;
+	return rbnode;
 }
 
 __weak
+u64 rb_node_alloc_internal(rbtree_t __arg_arena *rbtree, u64 key, u64 value)
+{
+	if (rbtree->alloc == RB_ALLOC)
+		return (u64)NULL;
+
+	return (u64)rb_node_alloc_common(rbtree, key, value);
+}
+
+__weak __attribute__((always_inline))
 int rb_node_free(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *rbnode)
 {
 	rbnode_t *old;
@@ -212,8 +223,8 @@ int rb_node_free(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *rbnode)
 	return 0;
 }
 
-__weak
-int rb_insert_node(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node, enum rbtree_insert_mode mode)
+static __attribute__((always_inline))
+int rb_node_insert(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node)
 {
 	rbnode_t *grandparent, *parent = rbtree->root;
 	u64 key = node->key;
@@ -229,13 +240,13 @@ int rb_insert_node(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node, enu
 		return 0;
 	}
 
-	if (mode != RB_DUPLICATE)
+	if (rbtree->insert != RB_DUPLICATE)
 		parent = rbnode_find(parent, key);
 	else
 		parent = rbnode_least_upper_bound(parent, key);
 
-	if (key == parent->key && mode != RB_DUPLICATE) {
-		if (mode == RB_UPDATE) {
+	if (key == parent->key && rbtree->insert != RB_DUPLICATE) {
+		if (rbtree->insert == RB_UPDATE) {
 			/*
 			 * Replace the old node with the new one.
 			 * Free up the old node.
@@ -245,7 +256,10 @@ int rb_insert_node(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node, enu
 			if (ret)
 				return ret;
 
-			rb_node_free(rbtree, parent);
+			/* Only free if called from rb_insert_node. */
+			if (rbtree->alloc == RB_ALLOC)
+				rb_node_free(rbtree, parent);
+
 			return 0;
 		}
 
@@ -303,19 +317,58 @@ int rb_insert_node(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node, enu
 	return 0;
 }
 
+int rb_insert_node(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node)
+{
+	volatile int i = 0;
+
+	if (unlikely(!rbtree))
+		return -EINVAL;
+
+	if (unlikely(rbtree->alloc == RB_ALLOC))
+		return -EINVAL;
+
+	node->is_red = true;
+	/* XXXETSAL: Variable i is not used. It is only there to 
+	 * prevent the compiler from causing verification failures 
+	 * in its attempt to optimize the series of assignments
+	 * to the rbnode_t * into a single operation.
+	 */
+
+	node->left = NULL;
+	i += 1;
+	node->right = NULL;
+	i += 1;
+	node->parent = NULL;
+
+	return rb_node_insert(rbtree, node);
+}
+
 __weak
-int rb_insert(rbtree_t __arg_arena *rbtree, u64 key, u64 value, enum rbtree_insert_mode mode)
+int rb_insert(rbtree_t __arg_arena *rbtree, u64 key, u64 value)
 {
 	rbnode_t *node;
+	int ret;
 
-	node = rb_node_alloc(rbtree, key, value);
+	if (unlikely(!rbtree))
+		return -EINVAL;
+
+	if (unlikely(rbtree->alloc != RB_ALLOC))
+		return -EINVAL;
+
+	node = rb_node_alloc_common(rbtree, key, value);
 	if (!node)
 		return -ENOMEM;
 
-	return rb_insert_node(rbtree, node, mode);
+	ret = rb_node_insert(rbtree, node);
+	if (ret) {
+		rb_node_free(rbtree, node);
+		return ret;
+	}
+
+	return 0;
 }
 
-static rbnode_t *rbnode_least(rbnode_t *subtree)
+static inline rbnode_t *rbnode_least(rbnode_t *subtree)
 {
 	while (subtree->left && can_loop)
 		subtree = subtree->left;
@@ -440,7 +493,7 @@ static void rbnode_switch(rbtree_t *rbtree, rbnode_t *a, rbnode_t *b)
 	rbnode_adjust_neighbors(rbtree, b, adir);
 }
 
-static inline int rbnode_remove_node_single_child(rbtree_t *rbtree, rbnode_t *node)
+static inline int rbnode_remove_node_single_child(rbtree_t *rbtree, rbnode_t *node, bool free)
 {
 	rbnode_t *child;
 	int dir;
@@ -472,7 +525,9 @@ static inline int rbnode_remove_node_single_child(rbtree_t *rbtree, rbnode_t *no
 	/* Color the child black. */
 	child->is_red = false;
 
-	rb_node_free(rbtree, node);
+	/* Only free if called from rb_remove. */
+	if (free)
+		rb_node_free(rbtree, node);
 
 	return 0;
 }
@@ -485,8 +540,8 @@ static inline bool rbnode_has_red_children(rbnode_t *node)
 	return node->right && node->right->is_red;
 }
 
-static
-int rb_remove_rbnode(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node)
+static __attribute__((always_inline))
+int rb_node_remove(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node, bool free)
 {
 	rbnode_t *parent, *sibling, *close_nephew, *distant_nephew;
 	rbnode_t *replace, *initial;
@@ -515,14 +570,15 @@ int rb_remove_rbnode(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node)
 
 	/* Only one child present, replace with child and paint it black. */
 	if (!node->left != !node->right)
-		return rbnode_remove_node_single_child(rbtree, node);
+		return rbnode_remove_node_single_child(rbtree, node, free);
 
 	/* (!node->left && !node->right) */
 
 	parent = node->parent;
 	if (!parent) {
 		rbtree->root = NULL;
-		rb_node_free(rbtree, node);
+		if (free)
+			rb_node_free(rbtree, node);
 		return 0;
 	}
 
@@ -530,7 +586,8 @@ int rb_remove_rbnode(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node)
 	parent->child[dir] = NULL;
 	is_red = node->is_red;
 
-	rb_node_free(rbtree, node);
+	if (free)
+		rb_node_free(rbtree, node);
 
 	/* If we removed a red node, we did not unbalance the tree.*/
 	if (is_red)
@@ -539,7 +596,6 @@ int rb_remove_rbnode(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node)
 	sibling = parent->child[1 - dir];
 	if (unlikely(!sibling)) {
 		bpf_printk("rbtree: removed black node has no sibling");
-		rb_print(rbtree);
 		return -EINVAL;
 	}
 
@@ -664,6 +720,18 @@ int rb_remove_rbnode(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node)
 	return 0;
 }
 
+__weak
+int rb_remove_node(rbtree_t __arg_arena *rbtree, rbnode_t __arg_arena *node)
+{
+	if (unlikely(!rbtree))
+		return -EINVAL;
+
+	if (unlikely(rbtree->alloc == RB_ALLOC))
+		return -EINVAL;
+
+	return rb_node_remove(rbtree, node, false);
+}
+
 __weak
 int rb_remove(rbtree_t __arg_arena *rbtree, u64 key)
 {
@@ -672,6 +740,9 @@ int rb_remove(rbtree_t __arg_arena *rbtree, u64 key)
 	if (unlikely(!rbtree))
 		return -EINVAL;
 
+	if (unlikely(rbtree->alloc != RB_ALLOC))
+		return -EINVAL;
+
 	if (!rbtree->root)
 		return -ENOENT;
 
@@ -679,7 +750,7 @@ int rb_remove(rbtree_t __arg_arena *rbtree, u64 key)
 	if (!node || node->key != key)
 		return -ENOENT;
 
-	return rb_remove_rbnode(rbtree, node);
+	return rb_node_remove(rbtree, node, true);
 }
 
 __weak
@@ -702,7 +773,7 @@ int rb_pop(rbtree_t __arg_arena *rbtree, u64 *key, u64 *value)
 	if (value)
 		*value = node->value;
 
-	return rb_remove_rbnode(rbtree, node);
+	return rb_remove_node(rbtree, node);
 }
 
 inline void rbnode_print(size_t depth, rbnode_t *rbn)
@@ -857,6 +928,9 @@ int rb_integrity_check(rbtree_t __arg_arena *rbtree)
 	if (unlikely(!rbtree))
 		return -EINVAL;
 
+	if (!rbtree->root)
+		return 0;
+
 	depth = 0;
 	state = RB_NONE_VISITED;
 
