@@ -11,8 +11,7 @@ pub mod bpf_intf;
 pub use bpf_intf::*;
 
 mod stats;
-use std::collections::BTreeMap;
-use std::ffi::c_int;
+use std::ffi::{c_int, c_ulong};
 use std::fmt::Write;
 use std::mem::MaybeUninit;
 use std::sync::atomic::AtomicBool;
@@ -20,6 +19,8 @@ use std::sync::atomic::Ordering;
 use std::sync::Arc;
 use std::time::Duration;
 
+use anyhow::anyhow;
+use anyhow::bail;
 use anyhow::Context;
 use anyhow::Result;
 use clap::Parser;
@@ -51,6 +52,7 @@ const SCHEDULER_NAME: &str = "scx_bpfland";
 
 #[derive(PartialEq)]
 enum Powermode {
+    Turbo,
     Performance,
     Powersave,
     Any,
@@ -118,23 +120,15 @@ struct Opts {
     exit_dump_len: u32,
 
     /// Maximum scheduling slice duration in microseconds.
-    #[clap(short = 's', long, default_value = "20000")]
+    #[clap(short = 's', long, default_value = "1000")]
     slice_us: u64,
 
-    /// Minimum scheduling slice duration in microseconds.
-    #[clap(short = 'S', long, default_value = "1000")]
-    slice_us_min: u64,
-
     /// Maximum time slice lag in microseconds.
     ///
     /// A positive value can help to enhance the responsiveness of interactive tasks, but it can
     /// also make performance more "spikey".
-    ///
-    /// A negative value can make performance more consistent, but it can also reduce the
-    /// responsiveness of interactive tasks (by smoothing the effect of the vruntime scheduling and
-    /// making the task ordering closer to a FIFO).
-    #[clap(short = 'l', long, allow_hyphen_values = true, default_value = "20000")]
-    slice_us_lag: i64,
+    #[clap(short = 'l', long, default_value = "40000")]
+    slice_us_lag: u64,
 
     /// Throttle the running CPUs by periodically injecting idle cycles.
     ///
@@ -151,13 +145,6 @@ struct Opts {
     #[clap(short = 'I', long, allow_hyphen_values = true, default_value = "-1")]
     idle_resume_us: i64,
 
-    /// Disable preemption.
-    ///
-    /// Never allow tasks to be directly dispatched. This can help to increase fairness
-    /// over responsiveness.
-    #[clap(short = 'n', long, action = clap::ArgAction::SetTrue)]
-    no_preempt: bool,
-
     /// Enable per-CPU tasks prioritization.
     ///
     /// This allows to prioritize per-CPU tasks that usually tend to be de-prioritized (since they
@@ -187,6 +174,16 @@ struct Opts {
     #[clap(short = 'w', long, action = clap::ArgAction::SetTrue)]
     no_wake_sync: bool,
 
+    /// Enable sticky tasks.
+    ///
+    /// If enabled force tasks with a high rate of enqueues/sec to stay on the same CPU, to reduce
+    /// locking contention on the shared runqueues.
+    ///
+    /// This can help making the scheduler more robust with intensive scheduling workloads and
+    /// benchmarks, but it can negatively impact on latency.
+    #[clap(short = 'S', long, action = clap::ArgAction::SetTrue)]
+    sticky_tasks: bool,
+
     /// Specifies the initial set of CPUs, represented as a bitmask in hex (e.g., 0xff), that the
     /// scheduler will use to dispatch tasks, until the system becomes saturated, at which point
     /// tasks may overflow to other available CPUs.
@@ -200,19 +197,18 @@ struct Opts {
     #[clap(short = 'm', long, default_value = "auto")]
     primary_domain: String,
 
-    /// Disable L2 cache awareness.
-    #[clap(long, action = clap::ArgAction::SetTrue)]
-    disable_l2: bool,
-
-    /// Disable L3 cache awareness.
-    #[clap(long, action = clap::ArgAction::SetTrue)]
-    disable_l3: bool,
+    /// Enable preferred idle CPU scanning.
+    ///
+    /// With this option enabled, the scheduler will prioritize assigning tasks to higher-ranked
+    /// cores before considering lower-ranked ones.
+    #[clap(short = 'P', long, action = clap::ArgAction::SetTrue)]
+    preferred_idle_scan: bool,
 
     /// Disable SMT awareness.
     #[clap(long, action = clap::ArgAction::SetTrue)]
     disable_smt: bool,
 
-    /// Disable NUMA rebalancing.
+    /// Disable NUMA awareness.
     #[clap(long, action = clap::ArgAction::SetTrue)]
     disable_numa: bool,
 
@@ -222,13 +218,6 @@ struct Opts {
     #[clap(short = 'f', long, action = clap::ArgAction::SetTrue)]
     cpufreq: bool,
 
-    /// [DEPRECATED] Maximum threshold of voluntary context switches per second. This is used to
-    /// classify interactive.
-    ///
-    /// tasks (0 = disable interactive tasks classification).
-    #[clap(short = 'c', long, default_value = "10", hide = true)]
-    nvcsw_max_thresh: u64,
-
     /// Enable stats monitoring with the specified interval.
     #[clap(long)]
     stats: Option<f64>,
@@ -272,9 +261,6 @@ impl<'a> Scheduler<'a> {
     fn init(opts: &'a Opts, open_object: &'a mut MaybeUninit<OpenObject>) -> Result<Self> {
         try_set_rlimit_infinity();
 
-        // Validate command line arguments.
-        assert!(opts.slice_us >= opts.slice_us_min);
-
         // Initialize CPU topology.
         let topo = Topology::new().unwrap();
 
@@ -290,11 +276,22 @@ impl<'a> Scheduler<'a> {
         info!("NUMA nodes: {}", nr_nodes);
 
         // Automatically disable NUMA optimizations when running on non-NUMA systems.
-        let numa_disabled = opts.disable_numa || nr_nodes == 1;
-        if numa_disabled {
+        let numa_enabled = !opts.disable_numa && nr_nodes > 1;
+        if !numa_enabled {
             info!("Disabling NUMA optimizations");
         }
 
+        // Determine the primary scheduling domain.
+        let power_profile = Self::power_profile();
+        let domain =
+            Self::resolve_energy_domain(&opts.primary_domain, power_profile).map_err(|err| {
+                anyhow!(
+                    "failed to resolve primary domain '{}': {}",
+                    &opts.primary_domain,
+                    err
+                )
+            })?;
+
         info!(
             "{} {} {}",
             SCHEDULER_NAME,
@@ -302,6 +299,12 @@ impl<'a> Scheduler<'a> {
             if smt_enabled { "SMT on" } else { "SMT off" }
         );
 
+        // Print command line.
+        info!(
+            "scheduler options: {}",
+            std::env::args().collect::<Vec<_>>().join(" ")
+        );
+
         if opts.idle_resume_us >= 0 {
             if !cpu_idle_resume_latency_supported() {
                 warn!("idle resume latency not supported");
@@ -328,31 +331,43 @@ impl<'a> Scheduler<'a> {
         let rodata = skel.maps.rodata_data.as_mut().unwrap();
         rodata.debug = opts.debug;
         rodata.smt_enabled = smt_enabled;
-        rodata.numa_disabled = numa_disabled;
+        rodata.numa_enabled = numa_enabled;
         rodata.local_pcpu = opts.local_pcpu;
-        rodata.no_preempt = opts.no_preempt;
         rodata.no_wake_sync = opts.no_wake_sync;
+        rodata.sticky_tasks = opts.sticky_tasks;
         rodata.slice_max = opts.slice_us * 1000;
-        rodata.slice_min = opts.slice_us_min * 1000;
         rodata.slice_lag = opts.slice_us_lag * 1000;
         rodata.throttle_ns = opts.throttle_us * 1000;
+        rodata.primary_all = domain.weight() == *NR_CPU_IDS;
+
+        // Generate the list of available CPUs sorted by capacity in descending order.
+        let mut cpus: Vec<_> = topo.all_cpus.values().collect();
+        cpus.sort_by_key(|cpu| std::cmp::Reverse(cpu.cpu_capacity));
+        for (i, cpu) in cpus.iter().enumerate() {
+            rodata.cpu_capacity[cpu.id] = cpu.cpu_capacity as c_ulong;
+            rodata.preferred_cpus[i] = cpu.id as u64;
+        }
+        if opts.preferred_idle_scan {
+            info!(
+                "Preferred CPUs: {:?}",
+                &rodata.preferred_cpus[0..cpus.len()]
+            );
+        }
+        rodata.preferred_idle_scan = opts.preferred_idle_scan;
 
         // Implicitly enable direct dispatch of per-CPU kthreads if CPU throttling is enabled
         // (it's never a good idea to throttle per-CPU kthreads).
         rodata.local_kthreads = opts.local_kthreads || opts.throttle_us > 0;
 
-        // Set scheduler compatibility flags.
-        rodata.__COMPAT_SCX_PICK_IDLE_IN_NODE = *compat::SCX_PICK_IDLE_IN_NODE;
-
         // Set scheduler flags.
         skel.struct_ops.bpfland_ops_mut().flags = *compat::SCX_OPS_ENQ_EXITING
             | *compat::SCX_OPS_ENQ_LAST
             | *compat::SCX_OPS_ENQ_MIGRATION_DISABLED
             | *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP
-            | if numa_disabled {
-                0
-            } else {
+            | if numa_enabled {
                 *compat::SCX_OPS_BUILTIN_IDLE_PER_NODE
+            } else {
+                0
             };
         info!(
             "scheduler flags: {:#x}",
@@ -362,13 +377,18 @@ impl<'a> Scheduler<'a> {
         // Load the BPF program for validation.
         let mut skel = scx_ops_load!(skel, bpfland_ops, uei)?;
 
-        // Initialize the primary scheduling domain and the preferred domain.
-        let power_profile = Self::power_profile();
-        if let Err(err) = Self::init_energy_domain(&mut skel, &opts.primary_domain, power_profile) {
-            warn!("failed to initialize primary domain: error {}", err);
-        }
+        // Initialize the primary scheduling domain.
+        Self::init_energy_domain(&mut skel, &domain).map_err(|err| {
+            anyhow!(
+                "failed to initialize primary domain 0x{:x}: {}",
+                domain,
+                err
+            )
+        })?;
+
+        // Initialize CPU frequency scaling.
         if let Err(err) = Self::init_cpufreq_perf(&mut skel, &opts.primary_domain, opts.cpufreq) {
-            warn!(
+            bail!(
                 "failed to initialize cpufreq performance level: error {}",
                 err
             );
@@ -379,15 +399,6 @@ impl<'a> Scheduler<'a> {
             Self::init_smt_domains(&mut skel, &topo)?;
         }
 
-        // Initialize L2 cache domains.
-        if !opts.disable_l2 {
-            Self::init_l2_cache_domains(&mut skel, &topo)?;
-        }
-        // Initialize L3 cache domains.
-        if !opts.disable_l3 {
-            Self::init_l3_cache_domains(&mut skel, &topo)?;
-        }
-
         // Attach the scheduler.
         let struct_ops = Some(scx_ops_attach!(skel, bpfland_ops)?);
         let stats_server = StatsServer::new(stats::server_data()).launch()?;
@@ -433,38 +444,37 @@ impl<'a> Scheduler<'a> {
         Cpumask::from_str(&cpus_to_cpumask(&cpus))
     }
 
-    fn init_energy_domain(
-        skel: &mut BpfSkel<'_>,
-        primary_domain: &str,
-        power_profile: PowerProfile,
-    ) -> Result<()> {
+    fn resolve_energy_domain(primary_domain: &str, power_profile: PowerProfile) -> Result<Cpumask> {
         let domain = match primary_domain {
             "powersave" => Self::epp_to_cpumask(Powermode::Powersave)?,
             "performance" => Self::epp_to_cpumask(Powermode::Performance)?,
+            "turbo" => Self::epp_to_cpumask(Powermode::Turbo)?,
             "auto" => match power_profile {
                 PowerProfile::Powersave => Self::epp_to_cpumask(Powermode::Powersave)?,
-                PowerProfile::Balanced { power: true } => {
-                    Self::epp_to_cpumask(Powermode::Powersave)?
-                }
-                PowerProfile::Balanced { power: false } => Self::epp_to_cpumask(Powermode::Any)?,
-                PowerProfile::Performance => Self::epp_to_cpumask(Powermode::Any)?,
-                PowerProfile::Unknown => Self::epp_to_cpumask(Powermode::Any)?,
+                PowerProfile::Balanced { .. }
+                | PowerProfile::Performance
+                | PowerProfile::Unknown => Self::epp_to_cpumask(Powermode::Any)?,
             },
             "all" => Self::epp_to_cpumask(Powermode::Any)?,
             &_ => Cpumask::from_str(primary_domain)?,
         };
 
+        Ok(domain)
+    }
+
+    fn init_energy_domain(skel: &mut BpfSkel<'_>, domain: &Cpumask) -> Result<()> {
         info!("primary CPU domain = 0x{:x}", domain);
 
         // Clear the primary domain by passing a negative CPU id.
         if let Err(err) = Self::enable_primary_cpu(skel, -1) {
-            warn!("failed to reset primary domain: error {}", err);
+            bail!("failed to reset primary domain: error {}", err);
         }
+
         // Update primary scheduling domain.
         for cpu in 0..*NR_CPU_IDS {
             if domain.test_cpu(cpu) {
                 if let Err(err) = Self::enable_primary_cpu(skel, cpu as i32) {
-                    warn!("failed to add CPU {} to primary domain: error {}", cpu, err);
+                    bail!("failed to add CPU {} to primary domain: error {}", cpu, err);
                 }
             }
         }
@@ -532,13 +542,11 @@ impl<'a> Scheduler<'a> {
 
     fn enable_sibling_cpu(
         skel: &mut BpfSkel<'_>,
-        lvl: usize,
         cpu: usize,
         sibling_cpu: usize,
     ) -> Result<(), u32> {
         let prog = &mut skel.progs.enable_sibling_cpu;
         let mut args = domain_arg {
-            lvl_id: lvl as c_int,
             cpu_id: cpu as c_int,
             sibling_cpu_id: sibling_cpu as c_int,
         };
@@ -564,96 +572,12 @@ impl<'a> Scheduler<'a> {
 
         info!("SMT sibling CPUs: {:?}", smt_siblings);
         for (cpu, sibling_cpu) in smt_siblings.iter().enumerate() {
-            Self::enable_sibling_cpu(skel, 0, cpu, *sibling_cpu as usize).unwrap();
+            Self::enable_sibling_cpu(skel, cpu, *sibling_cpu as usize).unwrap();
         }
 
         Ok(())
     }
 
-    fn are_smt_siblings(topo: &Topology, cpus: &[usize]) -> bool {
-        // Single CPU or empty array are considered siblings.
-        if cpus.len() <= 1 {
-            return true;
-        }
-
-        // Check if each CPU is a sibling of the first CPU.
-        let first_cpu = cpus[0];
-        let smt_siblings = topo.sibling_cpus();
-        cpus.iter().all(|&cpu| {
-            cpu == first_cpu
-                || smt_siblings[cpu] == first_cpu as i32
-                || (smt_siblings[first_cpu] >= 0 && smt_siblings[first_cpu] == cpu as i32)
-        })
-    }
-
-    fn init_cache_domains(
-        skel: &mut BpfSkel<'_>,
-        topo: &Topology,
-        cache_lvl: usize,
-        enable_sibling_cpu_fn: &dyn Fn(&mut BpfSkel<'_>, usize, usize, usize) -> Result<(), u32>,
-    ) -> Result<(), std::io::Error> {
-        // Determine the list of CPU IDs associated to each cache node.
-        let mut cache_id_map: BTreeMap<usize, Vec<usize>> = BTreeMap::new();
-        for core in topo.all_cores.values() {
-            for (cpu_id, cpu) in &core.cpus {
-                let cache_id = match cache_lvl {
-                    2 => cpu.l2_id,
-                    3 => cpu.llc_id,
-                    _ => panic!("invalid cache level {}", cache_lvl),
-                };
-                cache_id_map.entry(cache_id).or_default().push(*cpu_id);
-            }
-        }
-
-        // Update the BPF cpumasks for the cache domains.
-        for (cache_id, cpus) in cache_id_map {
-            // Ignore the cache domain if it includes a single CPU.
-            if cpus.len() <= 1 {
-                continue;
-            }
-
-            // Ignore the cache domain if all the CPUs are part of the same SMT core.
-            if Self::are_smt_siblings(topo, &cpus) {
-                continue;
-            }
-
-            info!(
-                "L{} cache ID {}: sibling CPUs: {:?}",
-                cache_lvl, cache_id, cpus
-            );
-            for cpu in &cpus {
-                for sibling_cpu in &cpus {
-                    if enable_sibling_cpu_fn(skel, cache_lvl, *cpu, *sibling_cpu).is_err() {
-                        warn!(
-                            "L{} cache ID {}: failed to set CPU {} sibling {}",
-                            cache_lvl, cache_id, *cpu, *sibling_cpu
-                        );
-                    }
-                }
-            }
-        }
-
-        Ok(())
-    }
-
-    fn init_l2_cache_domains(
-        skel: &mut BpfSkel<'_>,
-        topo: &Topology,
-    ) -> Result<(), std::io::Error> {
-        Self::init_cache_domains(skel, topo, 2, &|skel, lvl, cpu, sibling_cpu| {
-            Self::enable_sibling_cpu(skel, lvl, cpu, sibling_cpu)
-        })
-    }
-
-    fn init_l3_cache_domains(
-        skel: &mut BpfSkel<'_>,
-        topo: &Topology,
-    ) -> Result<(), std::io::Error> {
-        Self::init_cache_domains(skel, topo, 3, &|skel, lvl, cpu, sibling_cpu| {
-            Self::enable_sibling_cpu(skel, lvl, cpu, sibling_cpu)
-        })
-    }
-
     fn get_metrics(&self) -> Metrics {
         let bss_data = self.skel.maps.bss_data.as_ref().unwrap();
         Metrics {
