@@ -122,7 +122,7 @@ u64 get_suspended_duration_and_reset(struct cpu_ctx *cpuc)
 	return duration;
 }
 
-bool is_perf_cri(struct task_ctx *taskc)
+bool is_perf_cri(task_ctx __arg_arena *taskc)
 {
 	if (unlikely(!taskc))
 		return false;
@@ -131,7 +131,7 @@ bool is_perf_cri(struct task_ctx *taskc)
 		return true;
 
 	if (test_task_flag(taskc, LAVD_FLAG_ON_BIG | LAVD_FLAG_ON_LITTLE))
-		return taskc->perf_cri >= sys_stat.thr_perf_cri;
+		return taskc->perf_cri > sys_stat.thr_perf_cri;
 
 	return test_task_flag(taskc, LAVD_FLAG_ON_BIG);
 }
@@ -207,7 +207,7 @@ static int calc_nr_active_cpus(void)
 		else
 			WRITE_ONCE(pco_idx, nr_pco_states - 1);
 
-		bpf_for(i, 0, __nr_cpu_ids) {
+		bpf_for(i, 0, nr_cpu_ids) {
 			if (i >= LAVD_CPU_ID_MAX)
 				break;
 
@@ -239,7 +239,7 @@ static int calc_nr_active_cpus(void)
 	}
 
 	/* Should not be here. */
-	return __nr_cpu_ids;
+	return nr_cpu_ids;
 }
 
 __weak
@@ -277,7 +277,7 @@ int do_core_compaction(void)
 	/*
 	 * Assign active and overflow cores.
 	 */
-	bpf_for(i, 0, __nr_cpu_ids) {
+	bpf_for(i, 0, nr_cpu_ids) {
 		if (i >= LAVD_CPU_ID_MAX)
 			break;
 
@@ -312,11 +312,11 @@ int do_core_compaction(void)
 		} else {
 			bpf_cpumask_clear_cpu(cpu, active);
 
-			if (cpuc->nr_pinned_tasks || (per_cpu_dsq &&
+			if (cpuc->nr_pinned_tasks || (use_per_cpu_dsq() &&
 			    scx_bpf_dsq_nr_queued(cpu_to_dsq(cpu)))) {
 				/*
 				 * If there is something to run on this CPU,
-				 * add this CPU{ to the overflow set.
+				 * add this CPU to the overflow set.
 				 */
 				bpf_cpumask_set_cpu(cpu, ovrflw);
 			} else if ((bpf_get_prandom_u32() %
@@ -513,8 +513,14 @@ int update_thr_perf_cri(void)
 
 	/*
 	 * If all active cores are big, all tasks should run on the big cores.
+	 * On the other hand, if all active cores are small, all tasks should
+	 * run on the little cores.
 	 */
-	if (cur_big_core_scale == LAVD_SCALE) {
+	switch (cur_big_core_scale) {
+	case 0:
+		sys_stat.thr_perf_cri = sys_stat.max_perf_cri;
+		return 0;
+	case LAVD_SCALE:
 		sys_stat.thr_perf_cri = 0;
 		return 0;
 	}
@@ -547,16 +553,23 @@ int update_thr_perf_cri(void)
 		 *   |         |                       max_perf_cri
 		 *   |         |                       |
 		 *   <--------><----------------------->
-		 *
-		 *   <-///-><-------------------------->
-		 *   |     |                           |
-		 *   |     |                           1024
+		 *   |         \                       |
+		 *   |          \                      |
+		 *   |           \                     |
+		 *   |            \                    |
+		 *   |             \                   |
+		 *   |              \                  |
+		 *   <-///-><--------+----------------->
+		 *   |     |         |                 |
+		 *   |     |         512               1024
 		 *   |     little_core_scale
 		 *   0
+		 *
+		 * thr = min + (((avg - min) / 512) * little_core_scale)
 		 */
-		delta = (sys_stat.avg_perf_cri - sys_stat.min_perf_cri) >> 1;
-		diff = (delta * little_core_scale) >> LAVD_SHIFT;
-		thr = diff + sys_stat.min_perf_cri;
+		delta = (sys_stat.avg_perf_cri - sys_stat.min_perf_cri);
+		diff = (delta * little_core_scale) / (LAVD_SCALE >> 1);
+		thr = sys_stat.min_perf_cri + diff;
 	}
 	else {
 		/*
@@ -565,19 +578,26 @@ int update_thr_perf_cri(void)
 		 *   |         |                       max_perf_cri
 		 *   |         |                       |
 		 *   <--------><----------------------->
-		 *
-		 *   <---------------------><-////////->
-		 *   |                     |           |
-		 *   |                     |           1024
+		 *   |         \                       |
+		 *   |          \                      |
+		 *   |           \                     |
+		 *   |            \                    |
+		 *   |             \                   |
+		 *   |              \                  |
+		 *   <---------------+-----><-////////->
+		 *   |               |     |           |
+		 *   |               512   |           1024
 		 *   |                     little_core_scale
 		 *   0
 		 *
 		 *  Note that half of the little core capacity is taken by the
 		 *  [min_perf_cri, avg_perf_cri] range, so only another half
 		 *  can serve for the [avg_perf_cri, max_perf_cri range.
+		 *
+		 * thr = avg + (((max - avg) / 512) * (little_core_scale - 512))
 		 */
-		delta = (sys_stat.max_perf_cri - sys_stat.avg_perf_cri) >> 1;
-		diff = (delta * (little_core_scale >> 1)) >> LAVD_SHIFT;
+		delta = (sys_stat.max_perf_cri - sys_stat.avg_perf_cri);
+		diff = (delta * (little_core_scale - p2s(50))) / (LAVD_SCALE >> 1);
 		thr = diff + sys_stat.avg_perf_cri;
 	}
 
@@ -620,7 +640,7 @@ int reinit_active_cpumask_for_performance(void)
 	 * In a symmetric system, all online CPUs belong to the active set.
 	 */
 	if (have_little_core) {
-		bpf_for(cpu, 0, __nr_cpu_ids) {
+		bpf_for(cpu, 0, nr_cpu_ids) {
 			cpuc = get_cpu_ctx_id(cpu);
 			if (!cpuc)
 				continue;
@@ -653,7 +673,7 @@ int reinit_active_cpumask_for_performance(void)
 
 		bpf_cpumask_clear(ovrflw);
 
-		bpf_for(cpu, 0, __nr_cpu_ids) {
+		bpf_for(cpu, 0, nr_cpu_ids) {
 			cpuc = get_cpu_ctx_id(cpu);
 			if (!cpuc || !cpuc->is_online)
 				continue;
