@@ -34,6 +34,7 @@ volatile bool		no_freq_scaling;
 const volatile bool	no_wake_sync;
 const volatile bool	no_slice_boost;
 const volatile bool	per_cpu_dsq;
+const volatile bool	enable_cpu_bw;
 const volatile bool	is_autopilot_on;
 const volatile u8	verbose;
 
@@ -52,21 +53,10 @@ struct {
 	__uint(max_entries, 1);
 } cpu_ctx_stor SEC(".maps");
 
-/*
- * Per-task scheduling context
- */
-struct {
-	__uint(type, BPF_MAP_TYPE_TASK_STORAGE);
-	__uint(map_flags, BPF_F_NO_PREALLOC);
-	__type(key, int);
-	__type(value, struct task_ctx);
-} task_ctx_stor SEC(".maps");
-
-
 __hidden
-struct task_ctx *get_task_ctx(struct task_struct *p)
+u64 get_task_ctx_internal(struct task_struct __arg_trusted *p)
 {
-	return bpf_task_storage_get(&task_ctx_stor, p, 0, 0);
+	return (u64)scx_task_data(p);
 }
 
 __hidden
@@ -141,25 +131,30 @@ static bool is_kernel_worker(struct task_struct *p)
 	return !!(p->flags & (PF_WQ_WORKER | PF_IO_WORKER));
 }
 
+static bool is_ksoftirqd(struct task_struct *p)
+{
+	return is_kernel_task(p) && !__builtin_memcmp(p->comm, "ksoftirqd/", 10);
+}
+
 static bool is_pinned(const struct task_struct *p)
 {
 	return p->nr_cpus_allowed == 1;
 }
 
 __hidden
-bool test_task_flag(struct task_ctx *taskc, u64 flag)
+bool test_task_flag(task_ctx __arg_arena *taskc, u64 flag)
 {
 	return (taskc->flags & flag) == flag;
 }
 
 __hidden
-void set_task_flag(struct task_ctx *taskc, u64 flag)
+void set_task_flag(task_ctx __arg_arena *taskc, u64 flag)
 {
 	taskc->flags |= flag;
 }
 
 __hidden
-void reset_task_flag(struct task_ctx *taskc, u64 flag)
+void reset_task_flag(task_ctx __arg_arena *taskc, u64 flag)
 {
 	taskc->flags &= ~flag;
 }
@@ -183,13 +178,13 @@ inline void reset_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
 }
 
 __hidden
-bool is_lat_cri(struct task_ctx *taskc)
+bool is_lat_cri(task_ctx __arg_arena *taskc)
 {
 	return taskc->lat_cri >= sys_stat.avg_lat_cri;
 }
 
 __hidden
-bool is_lock_holder(struct task_ctx *taskc)
+bool is_lock_holder(task_ctx __arg_arena *taskc)
 {
 	return test_task_flag(taskc, LAVD_FLAG_FUTEX_BOOST);
 }
@@ -200,7 +195,7 @@ bool is_lock_holder_running(struct cpu_ctx *cpuc)
 	return test_cpu_flag(cpuc, LAVD_FLAG_FUTEX_BOOST);
 }
 
-bool have_scheduled(struct task_ctx *taskc)
+bool have_scheduled(task_ctx __arg_arena *taskc)
 {
 	/*
 	 * If task's time slice hasn't been updated, that means the task has
@@ -227,19 +222,26 @@ static bool use_full_cpus(void)
 
 s64 __attribute__ ((noinline)) pick_any_bit(u64 bitmap, u64 nuance)
 {
-	u64 i, pos;
+	u64 shift, rotated;
+	int tz;
 
-	#pragma clang loop unroll(disable)
-	bpf_for(i, 0, 64) {
-		pos = (i + nuance) % 64;
-		if (bitmap & (1LLU << pos))
-			return pos;
-	}
+	if (!bitmap)
+		return -ENOENT;
 
-	return -ENOENT;
+	/* modulo nuance to [0, 63] */
+	shift = nuance & 63ULL;
+
+	/* Circular rotate the bitmap by 'shift' bits. */
+	rotated = (bitmap >> shift) | (bitmap << (64 - shift));
+
+	/* Count the number of trailing zeros in the raomdonly rotated bitmap. */
+	tz = ctzll(rotated);
+
+	/* Add the shift back and wrap around to get the original index. */
+	return (tz + shift) & 63;
 }
 
-static void set_on_core_type(struct task_ctx *taskc,
+static void set_on_core_type(task_ctx __arg_arena *taskc,
 			     const struct cpumask *cpumask)
 {
 	bool on_big = false, on_little = false;
@@ -312,3 +314,25 @@ u32 cpu_to_dsq(u32 cpu)
 {
 	return (get_primary_cpu(cpu)) | LAVD_DSQ_TYPE_CPU << LAVD_DSQ_TYPE_SHFT;
 }
+
+s32 nr_queued_on_cpu(struct cpu_ctx *cpuc)
+{
+	s32 nr_queued;
+
+	nr_queued = scx_bpf_dsq_nr_queued(SCX_DSQ_LOCAL_ON | cpuc->cpu_id);
+
+	if (use_per_cpu_dsq())
+		nr_queued += scx_bpf_dsq_nr_queued(cpu_to_dsq(cpuc->cpu_id));
+
+	if (use_cpdom_dsq())
+		nr_queued += scx_bpf_dsq_nr_queued(cpdom_to_dsq(cpuc->cpdom_id));
+
+	return nr_queued;
+}
+
+u64 get_target_dsq_id(struct task_struct *p, struct cpu_ctx *cpuc)
+{
+	if (per_cpu_dsq || (pinned_slice_ns && is_pinned(p)))
+		return cpu_to_dsq(cpuc->cpu_id);
+	return cpdom_to_dsq(cpuc->cpdom_id);
+}
