@@ -14,6 +14,7 @@
 #include "../../../../include/scx/percpu.bpf.h"
 #include "../../../../include/lib/atq.h"
 #include "../../../../include/lib/cpumask.h"
+#include "../../../../include/lib/dhq.h"
 #include "../../../../include/lib/minheap.h"
 #include "../../../../include/lib/percpu.h"
 #include "../../../../include/lib/sdt_task.h"
@@ -24,6 +25,7 @@
 #include <scx/percpu.bpf.h>
 #include <lib/atq.h>
 #include <lib/cpumask.h>
+#include <lib/dhq.h>
 #include <lib/minheap.h>
 #include <lib/percpu.h>
 #include <lib/sdt_task.h>
@@ -117,14 +119,17 @@ const volatile struct {
 	u32 saturated_percent;
 	u32 sched_mode;
 	u32 llc_shards;
+	u64 dhq_max_imbalance;
 
 	bool atq_enabled;
+	bool dhq_enabled;
 	bool cpu_priority;
 	bool task_slice;
 	bool freq_control;
 	bool interactive_sticky;
 	bool keep_running_enabled;
 	bool kthreads_local;
+	bool pelt_enabled;
 } p2dq_config = {
 	.sched_mode = MODE_DEFAULT,
 	.nr_dsqs_per_llc = 3,
@@ -133,14 +138,26 @@ const volatile struct {
 	.interactive_ratio = 10,
 	.saturated_percent = 5,
 	.llc_shards = 0,
+	.dhq_max_imbalance = 3,
 
 	.atq_enabled = false,
+	.dhq_enabled = false,
 	.cpu_priority = false,
 	.task_slice = true,
 	.freq_control = false,
 	.interactive_sticky = false,
 	.keep_running_enabled = true,
 	.kthreads_local = true,
+	.pelt_enabled = true,
+};
+
+/* Latency priority and preemption configuration */
+const volatile struct {
+	bool latency_priority_enabled;
+	bool wakeup_preemption_enabled;
+} latency_config = {
+	.latency_priority_enabled = false,
+	.wakeup_preemption_enabled = false,
 };
 
 const volatile u32 debug = 2;
@@ -161,6 +178,13 @@ u64 cpu_node_ids[MAX_CPUS];
 u64 big_core_ids[MAX_CPUS];
 u64 dsq_time_slices[MAX_DSQS_PER_LLC];
 
+/* DHQ per LLC pair for migration (MAX_LLCS / 2 DHQs) */
+scx_dhq_t *llc_pair_dhqs[MAX_LLCS / 2];
+/* Track number of LLCs per NUMA node for strand assignment */
+u32 llcs_per_node[MAX_NUMA_NODES];
+/* Global DHQ counter for unique indexing */
+u32 global_dhq_count = 0;
+
 u64 min_slice_ns = 500;
 
 private(A) struct bpf_cpumask __kptr *all_cpumask;
@@ -273,6 +297,125 @@ static u32 nr_idle_cpus(const struct cpumask *idle_cpumask)
 	return nr_idle;
 }
 
+/*
+ * PELT (Per-Entity Load Tracking) helper functions
+ *
+ * Simplified BPF-friendly implementation of Linux kernel PELT.
+ * Uses 1ms periods and exponential decay with 32ms half-life.
+ */
+
+/*
+ * Apply exponential decay to a value over a number of periods.
+ * Each period decays by factor of 127/128 (≈ 0.98).
+ * Bounded loop for BPF verifier compliance.
+ */
+static __always_inline u32 pelt_decay(u32 val, u32 periods)
+{
+	u32 i;
+
+	/* Bound iterations for BPF verifier (max 256 periods = 256ms) */
+	bpf_for(i, 0, periods) {
+		if (i >= 256)
+			break;
+		val = (val * 127) >> 7;
+	}
+
+	return val;
+}
+
+/*
+ * Update task's PELT metrics based on runtime.
+ * Called when task stops running or starts running (for decay).
+ *
+ * @taskc: Task context to update
+ * @now: Current timestamp in ns
+ * @delta_ns: Runtime delta (0 for decay-only update)
+ */
+static __always_inline void update_task_pelt(task_ctx *taskc, u64 now, u64 delta_ns)
+{
+	u64 elapsed_ns, elapsed_ms;
+	u32 periods, delta_ms;
+
+	if (!p2dq_config.pelt_enabled)
+		return;
+
+	if (!taskc->pelt_last_update_time) {
+		/* First update - initialize */
+		taskc->pelt_last_update_time = now;
+		taskc->util_sum = 0;
+		taskc->util_avg = 0;
+		taskc->period_contrib = 0;
+		return;
+	}
+
+	elapsed_ns = now - taskc->pelt_last_update_time;
+	elapsed_ms = elapsed_ns / NSEC_PER_MSEC;
+
+	/**
+	 * If less than 1ms has passed, accumulate in period_contrib and don't
+	 * update timestamp until a full period has passed.
+	 */
+	if (elapsed_ms == 0) {
+		delta_ms = delta_ns / NSEC_PER_MSEC;
+		taskc->period_contrib += delta_ms;
+		return;
+	}
+
+	periods = (u32)elapsed_ms;
+	if (periods > 256)
+		periods = 256;  /* Cap for verifier */
+
+	if (taskc->util_sum > 0) {
+		taskc->util_sum = pelt_decay(taskc->util_sum, periods);
+	}
+
+	if (taskc->period_contrib > 0) {
+		taskc->util_sum += taskc->period_contrib;
+		taskc->period_contrib = 0;
+	}
+
+	delta_ms = delta_ns / NSEC_PER_MSEC;
+	taskc->util_sum += delta_ms;
+
+	if (unlikely(taskc->util_sum > PELT_SUM_MAX))
+		taskc->util_sum = PELT_SUM_MAX;
+
+	/* Calculate util_avg from util_sum */
+	/* util_avg = util_sum / 128 (representing average over ~128ms window) */
+	taskc->util_avg = taskc->util_sum >> 7;
+	if (taskc->util_avg > PELT_MAX_UTIL)
+		taskc->util_avg = PELT_MAX_UTIL;
+
+	taskc->pelt_last_update_time = now;
+}
+
+/*
+ * Aggregate task's PELT metrics to LLC context.
+ * Called when task stops running to update LLC utilization averages.
+ *
+ * @llcx: LLC context to update
+ * @taskc: Task context with updated PELT metrics
+ * @is_interactive: Whether task is interactive
+ * @is_affinitized: Whether task is affinitized to this LLC
+ */
+static __always_inline void aggregate_pelt_to_llc(struct llc_ctx *llcx,
+						   task_ctx *taskc,
+						   bool is_interactive,
+						   bool is_affinitized)
+{
+	if (!p2dq_config.pelt_enabled)
+		return;
+
+	__sync_fetch_and_add(&llcx->util_avg, taskc->util_avg);
+
+	if (is_interactive)
+		__sync_fetch_and_add(&llcx->intr_util_avg, taskc->util_avg);
+
+	if (is_affinitized)
+		__sync_fetch_and_add(&llcx->affn_util_avg, taskc->util_avg);
+}
+
+
 static u32 idle_cpu_percent(const struct cpumask *idle_cpumask)
 {
 	return (100 * nr_idle_cpus(idle_cpumask)) / topo_config.nr_cpus;
@@ -293,6 +436,14 @@ static void task_refresh_llc_runs(task_ctx *taskc)
 	taskc->llc_runs = min_llc_runs_pick2;
 }
 
+/*
+ * Get LLC load metric, using PELT util_avg if enabled, otherwise legacy load counter.
+ */
+static __always_inline u64 llc_get_load(const struct llc_ctx *llcx)
+{
+	return p2dq_config.pelt_enabled ? llcx->util_avg : llcx->load;
+}
+
 static u64 llc_nr_queued(struct llc_ctx *llcx)
 {
 	if (!llcx)
@@ -301,7 +452,9 @@ static u64 llc_nr_queued(struct llc_ctx *llcx)
 	u64 nr_queued = scx_bpf_dsq_nr_queued(llcx->dsq);
 
 	if (topo_config.nr_llcs > 1) {
-		if (p2dq_config.atq_enabled)
+		if (p2dq_config.dhq_enabled)
+			nr_queued += scx_dhq_nr_queued(llcx->mig_dhq);
+		else if (p2dq_config.atq_enabled)
 			nr_queued += scx_atq_nr_queued(llcx->mig_atq);
 		else
 			nr_queued += scx_bpf_dsq_nr_queued(llcx->mig_dsq);
@@ -330,6 +483,90 @@ static int llc_create_atqs(struct llc_ctx *llcx)
 	return 0;
 }
 
+/*
+ * Create DHQ for LLC pair migration.
+ * DHQs are shared between pairs of LLCs in the same NUMA node.
+ * Each LLC is assigned to a strand (A or B) based on its order in the node.
+ * Number of DHQs = number of LLCs per node / 2
+ */
+static int llc_create_dhqs(struct llc_ctx *llcx)
+{
+	u32 node_id = llcx->node_id;
+	u32 node_llc_count;
+	u32 dhq_index;
+	u64 strand;
+
+	if (!p2dq_config.dhq_enabled)
+		return 0;
+
+	if (topo_config.nr_llcs <= 1)
+		return 0;
+
+	if (node_id >= MAX_NUMA_NODES) {
+		scx_bpf_error("DHQ: node_id %u >= MAX_NUMA_NODES", node_id);
+		return -EINVAL;
+	}
+
+	/* Get current LLC count for this NUMA node */
+	node_llc_count = llcs_per_node[node_id];
+
+	/* Strand: A for first LLC in pair, B for second */
+	strand = (node_llc_count % 2 == 0) ? SCX_DHQ_STRAND_A : SCX_DHQ_STRAND_B;
+
+	/* First LLC in a pair: create a new DHQ */
+	if (strand == SCX_DHQ_STRAND_A) {
+		dhq_index = global_dhq_count;
+		if (dhq_index >= (MAX_LLCS / 2)) {
+			scx_bpf_error("DHQ: dhq_index %u >= MAX_LLCS/2", dhq_index);
+			return -EINVAL;
+		}
+
+		/* Create fixed-size DHQ with priority mode for lowest vtime selection.
+		 * Capacity scales with system size: 4x CPUs ensures enough headroom
+		 * for queued tasks under load without excessive memory usage.
+		 * Max imbalance controls strand balance for cross-LLC load balancing.
+		 */
+		u64 dhq_capacity = topo_config.nr_cpus * 4;
+		llc_pair_dhqs[dhq_index] = (scx_dhq_t *)scx_dhq_create_balanced(
+			false,                          /* vtime mode */
+			dhq_capacity,                   /* fixed capacity */
+			SCX_DHQ_MODE_PRIORITY,          /* lowest vtime wins */
+			p2dq_config.dhq_max_imbalance   /* max_imbalance from config */
+		);
+		if (!llc_pair_dhqs[dhq_index]) {
+			scx_bpf_error("DHQ failed to create DHQ %u for node %u",
+				      dhq_index, node_id);
+			return -ENOMEM;
+		}
+		trace("DHQ %u created for node %u (LLC %u, strand A) capacity=%llu",
+		      dhq_index, node_id, llcx->id, dhq_capacity);
+
+		/* Assign DHQ and strand to this LLC */
+		llcx->mig_dhq = llc_pair_dhqs[dhq_index];
+		llcx->dhq_strand = strand;
+		global_dhq_count++;
+	} else {
+		/* Second LLC in pair: use the most recently created DHQ */
+		dhq_index = global_dhq_count - 1;
+		if (dhq_index >= (MAX_LLCS / 2) || !llc_pair_dhqs[dhq_index]) {
+			scx_bpf_error("DHQ: DHQ %u not available for second LLC %u",
+				      dhq_index, llcx->id);
+			return -EINVAL;
+		}
+		trace("DHQ %u assigned to LLC %u (node %u, strand B)",
+		      dhq_index, llcx->id, node_id);
+
+		/* Assign DHQ and strand to this LLC */
+		llcx->mig_dhq = llc_pair_dhqs[dhq_index];
+		llcx->dhq_strand = strand;
+	}
+
+	llcs_per_node[node_id]++;
+
+	return 0;
+}
+
+
 struct p2dq_timer p2dq_timers[MAX_TIMERS] = {
 	{lb_timer_intvl_ns,
 	     CLOCK_BOOTTIME, 0},
@@ -594,6 +831,12 @@ static s32 pick_idle_affinitized_cpu(struct task_struct *p, task_ctx *taskc,
 	struct llc_ctx *llcx;
 	s32 cpu = prev_cpu;
 
+	// Migration-disabled tasks must stay on their current CPU
+	if (is_migration_disabled(p)) {
+		*is_idle = scx_bpf_test_and_clear_cpu_idle(prev_cpu);
+		return prev_cpu;
+	}
+
 	idle_cpumask = scx_bpf_get_idle_cpumask();
 	idle_smtmask = scx_bpf_get_idle_smtmask();
 
@@ -673,19 +916,19 @@ static s32 pick_idle_cpu(struct task_struct *p, task_ctx *taskc,
 	if (!idle_cpumask || !idle_smtmask)
 		goto found_cpu;
 
+	if (bpf_cpumask_test_cpu(prev_cpu, idle_cpumask) &&
+	    scx_bpf_test_and_clear_cpu_idle(prev_cpu)) {
+		*is_idle = true;
+		goto found_cpu;
+	}
+
 	if (p2dq_config.interactive_sticky && task_ctx_test_flag(taskc, TASK_CTX_F_INTERACTIVE)) {
 		*is_idle = scx_bpf_test_and_clear_cpu_idle(prev_cpu);
 		goto found_cpu;
 	}
 
-	// First check if last CPU is idle (common case - CPU cache warm)
-	if (likely(bpf_cpumask_test_cpu(prev_cpu,
-				 (topo_config.smt_enabled && !task_ctx_test_flag(taskc, TASK_CTX_F_INTERACTIVE)) ?
-				 idle_smtmask : idle_cpumask) &&
-	    scx_bpf_test_and_clear_cpu_idle(prev_cpu))) {
-		*is_idle = true;
+	if (idle_cpumask && bpf_cpumask_empty(idle_cpumask))
 		goto found_cpu;
-	}
 
 	if (!(llcx = lookup_llc_ctx(taskc->llc_id)) ||
 	    !llcx->cpumask)
@@ -943,6 +1186,39 @@ static s32 p2dq_select_cpu_impl(struct task_struct *p, s32 prev_cpu, u64 wake_fl
 	else
 		cpu = pick_idle_cpu(p, taskc, prev_cpu, wake_flags, &is_idle);
 
+	// Wakeup preemption for extremely latency-critical tasks
+	// Only attempt if: no idle CPU found AND task has very high priority
+	if (!is_idle && latency_config.wakeup_preemption_enabled) {
+		struct cpu_ctx *prev_cpuc;
+
+		// Only preempt for truly latency-critical tasks (scx.weight >= 2847, equivalent to nice <= -15)
+		// and only if we can check the prev_cpu state
+		if (p->scx.weight >= 2847 &&
+		    bpf_cpumask_test_cpu(prev_cpu, p->cpus_ptr) &&
+		    (prev_cpuc = lookup_cpu_ctx(prev_cpu))) {
+
+			// Don't preempt interactive tasks - they need low latency too
+			if (cpu_ctx_test_flag(prev_cpuc, CPU_CTX_F_INTERACTIVE)) {
+				goto skip_preempt;
+			}
+
+			// Only preempt if incoming task has higher priority than running task
+			// This ensures we only preempt lower priority work
+			if (p->scx.weight <= prev_cpuc->running_weight) {
+				goto skip_preempt;
+			}
+
+			// Queue to prev_cpu's LLC DSQ with high priority
+			// Don't bypass normal queueing - let vtime ordering work
+			// Just ensure we target prev_cpu for better cache affinity
+			cpu = prev_cpu;
+			trace("PREEMPT_TARGET [%d][%s] weight=%u > running_weight=%u on cpu=%d",
+			      p->pid, p->comm, p->scx.weight, prev_cpuc->running_weight, prev_cpu);
+		}
+	}
+
+skip_preempt:
+
 	if (likely(is_idle)) {
 		stat_inc(P2DQ_STAT_IDLE);
 		// Only direct dispatch non-affinitized tasks
@@ -1008,11 +1284,9 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 
 	// Handle affinitized tasks: always use per-CPU affn_dsq
 	// All affinitized tasks queued to affn_dsq regardless of affinity breadth
-	// Also handle interactive tasks when interactive_sticky is enabled
 	if (!task_ctx_test_flag(taskc, TASK_CTX_F_ALL_CPUS) ||
 	    (p->cpus_ptr == &p->cpus_mask &&
-	     p->nr_cpus_allowed != topo_config.nr_cpus) ||
-	    (p2dq_config.interactive_sticky && task_ctx_test_flag(taskc, TASK_CTX_F_INTERACTIVE))) {
+	     p->nr_cpus_allowed != topo_config.nr_cpus)) {
 		bool has_cleared_idle = false;
 		if (!__COMPAT_is_enq_cpu_selected(enq_flags) ||
 		    !bpf_cpumask_test_cpu(cpu, p->cpus_ptr))
@@ -1035,11 +1309,7 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 			return;
 		}
 
-		// Track whether this is an interactive sticky enqueue or regular affinity enqueue
-		if (p2dq_config.interactive_sticky && task_ctx_test_flag(taskc, TASK_CTX_F_INTERACTIVE))
-			stat_inc(P2DQ_STAT_ENQ_INTR);
-		else
-			stat_inc(P2DQ_STAT_ENQ_CPU);
+		stat_inc(P2DQ_STAT_ENQ_CPU);
 
 		// Select target CPU for affn_dsq with priority:
 		// 1. prev_cpu if in affinity
@@ -1110,11 +1380,13 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 
 		// Always queue affinitized tasks to affn_dsq (no direct dispatch)
 		// This prevents tight wakeup loops and allows proper idle state
+		u64 task_vtime_affn = p->scx.dsq_vtime;
+
 		ret->kind = P2DQ_ENQUEUE_PROMISE_VTIME;
 		ret->vtime.dsq_id = taskc->dsq_id;
 		ret->vtime.slice_ns = taskc->slice_ns;
 		ret->vtime.enq_flags = enq_flags;
-		ret->vtime.vtime = p->scx.dsq_vtime;
+		ret->vtime.vtime = task_vtime_affn;
 
 		// Kick target CPU if we cleared idle state
 		if (enqueue_promise_test_flag(ret, ENQUEUE_PROMISE_F_HAS_CLEARED_IDLE))
@@ -1122,7 +1394,7 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 
 		trace("ENQUEUE %s weight %d slice %llu vtime %llu llc vtime %llu affn_dsq",
 		      p->comm, p->scx.weight, taskc->slice_ns,
-		      p->scx.dsq_vtime, llcx->vtime);
+		      task_vtime_affn, llcx->vtime);
 
 		return;
 	}
@@ -1146,6 +1418,7 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 			return;
 		}
 
+		s32 task_cpu = scx_bpf_task_cpu(p);
 		ret->cpu = cpu;
 		update_vtime(p, cpuc, taskc, llcx);
 		if (timeline_config.deadline)
@@ -1155,10 +1428,15 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 			enq_flags |= SCX_ENQ_PREEMPT;
 
 		if ((enqueue_promise_test_flag(ret, ENQUEUE_PROMISE_F_HAS_CLEARED_IDLE) ||
-		     cpu_ctx_test_flag(cpuc, CPU_CTX_F_NICE_TASK)) &&
-		    bpf_cpumask_test_cpu(cpu, p->cpus_ptr)) {
+		     cpu_ctx_test_flag(cpuc, CPU_CTX_F_NICE_TASK))) {
 			ret->kind = P2DQ_ENQUEUE_PROMISE_FIFO;
-			ret->fifo.dsq_id = SCX_DSQ_LOCAL_ON|cpu;
+			// For migration-disabled tasks, use SCX_DSQ_LOCAL to dispatch
+			// to the task's current CPU, not SCX_DSQ_LOCAL_ON|cpu
+			if (cpu != task_cpu && !is_migration_disabled(p)) {
+				ret->fifo.dsq_id = SCX_DSQ_LOCAL_ON|cpu;
+			} else {
+				ret->fifo.dsq_id = SCX_DSQ_LOCAL;
+			}
 			ret->fifo.slice_ns = taskc->slice_ns;
 			ret->fifo.enq_flags = enq_flags;
 			if (enqueue_promise_test_flag(ret, ENQUEUE_PROMISE_F_HAS_CLEARED_IDLE))
@@ -1171,21 +1449,33 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 		bool migrate = likely(!lb_config.single_llc_mode) &&
 		               can_migrate(taskc, llcx) &&
 		               task_ctx_test_flag(taskc, TASK_CTX_F_ALL_CPUS);
+
+		u64 task_vtime_early = p->scx.dsq_vtime;
+
 		if (migrate) {
 			taskc->dsq_id = llcx->mig_dsq;
-			if (p2dq_config.atq_enabled) {
+			if (p2dq_config.dhq_enabled) {
+				taskc->enq_flags = enq_flags;
+				ret->kind = P2DQ_ENQUEUE_PROMISE_DHQ_VTIME;
+				ret->dhq.dsq_id = cpuc->llc_dsq;
+				ret->dhq.dhq = llcx->mig_dhq;
+				ret->dhq.strand = llcx->dhq_strand;
+				ret->dhq.slice_ns = taskc->slice_ns;
+				ret->dhq.vtime = task_vtime_early;
+				ret->dhq.enq_flags = enq_flags;
+			} else if (p2dq_config.atq_enabled) {
 				taskc->enq_flags = enq_flags;
 				ret->kind = P2DQ_ENQUEUE_PROMISE_ATQ_VTIME;
 				ret->vtime.dsq_id = cpuc->llc_dsq;
 				ret->vtime.atq = llcx->mig_atq;
 				ret->vtime.slice_ns = taskc->slice_ns;
-				ret->vtime.vtime = p->scx.dsq_vtime;
+				ret->vtime.vtime = task_vtime_early;
 			} else {
 				ret->kind = P2DQ_ENQUEUE_PROMISE_VTIME;
 				ret->vtime.dsq_id = taskc->dsq_id;
 				ret->vtime.slice_ns = taskc->slice_ns;
 				ret->vtime.enq_flags = enq_flags;
-				ret->vtime.vtime = p->scx.dsq_vtime;
+				ret->vtime.vtime = task_vtime_early;
 			}
 			stat_inc(P2DQ_STAT_ENQ_MIG);
 		} else {
@@ -1194,13 +1484,13 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 			ret->vtime.dsq_id = taskc->dsq_id;
 			ret->vtime.slice_ns = taskc->slice_ns;
 			ret->vtime.enq_flags = enq_flags;
-			ret->vtime.vtime = p->scx.dsq_vtime;
+			ret->vtime.vtime = task_vtime_early;
 			stat_inc(P2DQ_STAT_ENQ_LLC);
 		}
 
 		trace("ENQUEUE %s weight %d slice %llu vtime %llu llc vtime %llu",
 		      p->comm, p->scx.weight, taskc->slice_ns,
-		      p->scx.dsq_vtime, llcx->vtime);
+		      task_vtime_early, llcx->vtime);
 
 		return;
 	}
@@ -1245,13 +1535,27 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 	if (migrate) {
 		taskc->dsq_id = llcx->mig_dsq;
 		stat_inc(P2DQ_STAT_ENQ_MIG);
-		if (p2dq_config.atq_enabled) {
+
+		u64 task_vtime_mig = p->scx.dsq_vtime;
+
+		if (p2dq_config.dhq_enabled) {
+			taskc->enq_flags = enq_flags;
+			ret->kind = P2DQ_ENQUEUE_PROMISE_DHQ_VTIME;
+			ret->dhq.dsq_id = cpuc->llc_dsq;
+			ret->dhq.dhq = llcx->mig_dhq;
+			ret->dhq.strand = llcx->dhq_strand;
+			ret->dhq.slice_ns = taskc->slice_ns;
+			ret->dhq.vtime = task_vtime_mig;
+			ret->dhq.enq_flags = enq_flags;
+
+			return;
+		} else if (p2dq_config.atq_enabled) {
 			taskc->enq_flags = enq_flags;
 			ret->kind = P2DQ_ENQUEUE_PROMISE_ATQ_VTIME;
 			ret->vtime.dsq_id = cpuc->llc_dsq;
 			ret->vtime.atq = llcx->mig_atq;
 			ret->vtime.slice_ns = taskc->slice_ns;
-			ret->vtime.vtime = p->scx.dsq_vtime;
+			ret->vtime.vtime = task_vtime_mig;
 
 			return;
 		}
@@ -1260,19 +1564,22 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 		stat_inc(P2DQ_STAT_ENQ_LLC);
 	}
 
+	u64 task_vtime = p->scx.dsq_vtime;
+
 	trace("ENQUEUE %s weight %d slice %llu vtime %llu llc vtime %llu",
 	      p->comm, p->scx.weight, taskc->slice_ns,
-	      p->scx.dsq_vtime, llcx->vtime);
+	      task_vtime, llcx->vtime);
 
 	ret->kind = P2DQ_ENQUEUE_PROMISE_VTIME;
 	ret->vtime.dsq_id = taskc->dsq_id;
 	ret->vtime.enq_flags = enq_flags;
 	ret->vtime.slice_ns = taskc->slice_ns;
-	ret->vtime.vtime = p->scx.dsq_vtime;
+	ret->vtime.vtime = task_vtime;
 }
 
 static void complete_p2dq_enqueue(struct enqueue_promise *pro, struct task_struct *p)
 {
+	task_ctx *taskc;
 	int ret;
 
 	switch (pro->kind) {
@@ -1293,36 +1600,52 @@ static void complete_p2dq_enqueue(struct enqueue_promise *pro, struct task_struc
 		break;
 	case P2DQ_ENQUEUE_PROMISE_ATQ_FIFO:
 		if (!pro->fifo.atq) {
-			scx_bpf_error("invalid ATQ");
+			scx_bpf_error("promise has no fifo ATQ");
 			break;
 		}
-		ret = scx_atq_insert(pro->fifo.atq, (u64)p->pid);
-		if (!ret) {
-			// The ATQ was full, fallback to the DSQ.
-			scx_bpf_dsq_insert(p,
-					   pro->vtime.dsq_id,
-					   pro->vtime.slice_ns,
-					   pro->vtime.enq_flags);
-			stat_inc(P2DQ_STAT_ATQ_REENQ);
-		} else {
-			stat_inc(P2DQ_STAT_ATQ_ENQ);
+
+		taskc = lookup_task_ctx(p);
+		ret = scx_atq_insert(pro->fifo.atq, &taskc->common);
+		if (ret) {
+			scx_bpf_error("error %d on scx_atq_insert", ret);
+			break;
 		}
+
+		stat_inc(P2DQ_STAT_ATQ_ENQ);
 		break;
 	case P2DQ_ENQUEUE_PROMISE_ATQ_VTIME:
+
 		if (!pro->vtime.atq) {
-			scx_bpf_error("invalid ATQ");
+			scx_bpf_error("promise has no vtime ATQ");
 			break;
 		}
+
+		taskc = lookup_task_ctx(p);
 		ret = scx_atq_insert_vtime(pro->vtime.atq,
-					       (u64)p->pid,
+					       &taskc->common,
 					       pro->vtime.vtime);
-		if (!ret) {
-			// The ATQ was full, fallback to the DSQ.
+		if (ret) {
+			scx_bpf_error("error %d on scx_atq_insert", ret);
+			break;
+		}
+		break;
+	case P2DQ_ENQUEUE_PROMISE_DHQ_VTIME:
+		if (!pro->dhq.dhq) {
+			scx_bpf_error("invalid DHQ");
+			break;
+		}
+		ret = scx_dhq_insert_vtime(pro->dhq.dhq,
+					   (u64)p->pid,
+					   pro->dhq.vtime,
+					   pro->dhq.strand);
+		if (ret) {
+			// The DHQ insert failed (EAGAIN if imbalanced, ENOSPC if full)
+			// Fallback to the DSQ
 			scx_bpf_dsq_insert_vtime(p,
-						 pro->vtime.dsq_id,
-						 pro->vtime.slice_ns,
-						 pro->vtime.vtime,
-						 pro->vtime.enq_flags);
+						 pro->dhq.dsq_id,
+						 pro->dhq.slice_ns,
+						 pro->dhq.vtime,
+						 pro->dhq.enq_flags);
 			stat_inc(P2DQ_STAT_ATQ_REENQ);
 		} else {
 			stat_inc(P2DQ_STAT_ATQ_ENQ);
@@ -1384,6 +1707,7 @@ static int p2dq_running_impl(struct task_struct *p)
 		cpu_ctx_clear_flag(cpuc, CPU_CTX_F_INTERACTIVE);
 
 	cpuc->dsq_index = taskc->dsq_index;
+	cpuc->running_weight = p->scx.weight;
 
 	if (p->scx.weight < 100)
 		cpu_ctx_set_flag(cpuc, CPU_CTX_F_NICE_TASK);
@@ -1412,6 +1736,10 @@ static int p2dq_running_impl(struct task_struct *p)
 
 	taskc->last_run_at = now;
 
+	/* Decay PELT metrics when task starts running (0 delta for decay-only) */
+	if (p2dq_config.pelt_enabled)
+		update_task_pelt(taskc, now, 0);
+
 	return 0;
 }
 
@@ -1453,16 +1781,28 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 
 	p->scx.dsq_vtime += scaled_used;
 	__sync_fetch_and_add(&llcx->vtime, used);
-	__sync_fetch_and_add(&llcx->load, used);
-	if (taskc->dsq_index >= 0 && taskc->dsq_index < MAX_DSQS_PER_LLC)
-		__sync_fetch_and_add(&llcx->dsq_load[taskc->dsq_index], used);
 
-	if (task_ctx_test_flag(taskc, TASK_CTX_F_INTERACTIVE))
-		__sync_fetch_and_add(&llcx->intr_load, used);
+	/* Update PELT metrics if enabled */
+	if (p2dq_config.pelt_enabled) {
+		update_task_pelt(taskc, now, used);
+		aggregate_pelt_to_llc(llcx, taskc,
+				      task_ctx_test_flag(taskc, TASK_CTX_F_INTERACTIVE),
+				      !task_ctx_test_flag(taskc, TASK_CTX_F_ALL_CPUS));
+	}
 
-	if (!task_ctx_test_flag(taskc, TASK_CTX_F_ALL_CPUS))
-		// Note that affinitized load is absolute load, not scaled.
-		__sync_fetch_and_add(&llcx->affn_load, used);
+	/* Legacy load tracking (when PELT disabled) */
+	if (!p2dq_config.pelt_enabled) {
+		__sync_fetch_and_add(&llcx->load, used);
+		if (taskc->dsq_index >= 0 && taskc->dsq_index < MAX_DSQS_PER_LLC)
+			__sync_fetch_and_add(&llcx->dsq_load[taskc->dsq_index], used);
+
+		if (task_ctx_test_flag(taskc, TASK_CTX_F_INTERACTIVE))
+			__sync_fetch_and_add(&llcx->intr_load, used);
+
+		if (!task_ctx_test_flag(taskc, TASK_CTX_F_ALL_CPUS))
+			// Note that affinitized load is absolute load, not scaled.
+			__sync_fetch_and_add(&llcx->affn_load, used);
+	}
 
 	trace("STOPPING %s weight %d slice %llu used %llu scaled %llu",
 	      p->comm, p->scx.weight, last_dsq_slice_ns, used, scaled_used);
@@ -1532,36 +1872,74 @@ static bool consume_llc(struct llc_ctx *llcx)
 {
 	struct task_struct *p;
 	task_ctx *taskc;
+	struct cpu_ctx *cpuc;
+	s32 cpu;
 	u64 pid;
 
 	if (!llcx)
 		return false;
 
-	if (p2dq_config.atq_enabled &&
-	    scx_atq_nr_queued(llcx->mig_atq) > 0) {
-		pid = scx_atq_pop(llcx->mig_atq);
+	cpu = bpf_get_smp_processor_id();
+	if (!(cpuc = lookup_cpu_ctx(cpu)))
+		return false;
+
+	if (p2dq_config.dhq_enabled &&
+	    scx_dhq_nr_queued(llcx->mig_dhq) > 0) {
+		pid = scx_dhq_pop_strand(llcx->mig_dhq, llcx->dhq_strand);
+		if (!pid) {
+			trace("DHQ pop returned NULL");
+			goto try_dsq;
+		}
+
 		p = bpf_task_from_pid((s32)pid);
 		if (!p) {
-			trace("ATQ failed to get pid %llu", pid);
-			return false;
+			trace("DHQ failed to get pid %llu", pid);
+			goto try_dsq;
 		}
 
 		if (!(taskc = lookup_task_ctx(p))) {
 			bpf_task_release(p);
+			goto try_dsq;
+		}
+
+		/* Insert to LLC DSQ and let move_to_local handle affinity atomically */
+		trace("DHQ %llu insert %s[%d] to LLC DSQ",
+		      llcx->mig_dhq, p->comm, p->pid);
+		scx_bpf_dsq_insert_vtime(p,
+					 cpuc->llc_dsq,
+					 taskc->slice_ns,
+					 p->scx.dsq_vtime,
+					 taskc->enq_flags);
+		bpf_task_release(p);
+
+		/* Try to dispatch from LLC DSQ (handles affinity check atomically) */
+		if (scx_bpf_dsq_move_to_local(cpuc->llc_dsq))
+			return true;
+
+		goto try_dsq;
+	} else if (p2dq_config.atq_enabled &&
+	    scx_atq_nr_queued(llcx->mig_atq) > 0) {
+		taskc = (task_ctx *)scx_atq_pop(llcx->mig_atq);
+		p = bpf_task_from_pid((s32)taskc->pid);
+		if (!p) {
+			trace("ATQ failed to get pid %llu", taskc->pid);
 			return false;
 		}
 
-		trace("ATQ %llu insert %s->%d",
+/* Insert to LLC DSQ and let move_to_local handle affinity atomically */
+		trace("ATQ %llu insert %s[%d] to LLC DSQ",
 		      llcx->mig_atq, p->comm, p->pid);
-		scx_bpf_dsq_insert(p,
-				   SCX_DSQ_LOCAL,
-				   taskc->slice_ns,
-				   taskc->enq_flags);
+		scx_bpf_dsq_insert_vtime(p,
+					 cpuc->llc_dsq,
+					 taskc->slice_ns,
+					 p->scx.dsq_vtime,
+					 taskc->enq_flags);
 		bpf_task_release(p);
 
-		return false;
+		/* Try to dispatch from LLC DSQ (handles affinity check atomically) */
+		return scx_bpf_dsq_move_to_local(cpuc->llc_dsq);
 	}
-
+try_dsq:
 	if (likely(scx_bpf_dsq_move_to_local(llcx->mig_dsq))) {
 		stat_inc(P2DQ_STAT_DISPATCH_PICK2);
 		return true;
@@ -1616,7 +1994,7 @@ static __always_inline int dispatch_pick_two(s32 cpu, struct llc_ctx *cur_llcx,
 		return -EINVAL;
 
 	if (left->id == right->id) {
-		i = cur_llcx->load % topo_config.nr_llcs;
+		i = llc_get_load(cur_llcx) % topo_config.nr_llcs;
 		i &= 0x3; // verifier
 		if (i >= 0 && i < topo_config.nr_llcs)
 			right = lookup_llc_ctx(llc_ids[i]);
@@ -1625,7 +2003,7 @@ static __always_inline int dispatch_pick_two(s32 cpu, struct llc_ctx *cur_llcx,
 	}
 
 
-	if (right->load > left->load) {
+	if (llc_get_load(right) > llc_get_load(left)) {
 		first = right;
 		second = left;
 	} else {
@@ -1641,15 +2019,15 @@ static __always_inline int dispatch_pick_two(s32 cpu, struct llc_ctx *cur_llcx,
 	}
 
 	trace("PICK2 cpu[%d] first[%d] %llu second[%d] %llu",
-	      cpu, first->id, first->load, second->id, second->load);
+	      cpu, first->id, llc_get_load(first), second->id, llc_get_load(second));
 
-	cur_load = cur_llcx->load + ((cur_llcx->load * lb_config.slack_factor) / 100);
+	cur_load = llc_get_load(cur_llcx) + ((llc_get_load(cur_llcx) * lb_config.slack_factor) / 100);
 
-	if (first->load >= cur_load &&
+	if (llc_get_load(first) >= cur_load &&
 	    consume_llc(first))
 		return 0;
 
-	if (second->load >= cur_load &&
+	if (llc_get_load(second) >= cur_load &&
 	    consume_llc(second))
 		return 0;
 
@@ -1677,8 +2055,9 @@ static void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 	task_ctx *taskc;
 	struct cpu_ctx *cpuc;
 	struct llc_ctx *llcx;
-	u64 pid, peeked_pid, dsq_id = 0;
+	u64 pid, dsq_id = 0;
 	scx_atq_t *min_atq = NULL;
+	scx_dhq_t *min_dhq = NULL;
 
 	cpuc = lookup_cpu_ctx(cpu);
 	if (unlikely(!cpuc)) {
@@ -1688,6 +2067,8 @@ static void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 
 	u64 min_vtime = 0;
 
+	bpf_rcu_read_lock();
+
 	// start with affn_dsq (local cpu dsq)
 	p = __COMPAT_scx_bpf_dsq_peek(cpuc->affn_dsq);
 	if (p) {
@@ -1731,6 +2112,42 @@ static void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 		}
 	}
 
+	// Check other CPUs' affn_dsq in same LLC for affinitized work stealing
+	// This prevents high wakeup latency when tasks are queued on busy CPUs
+	// but other CPUs in the affinity mask are idle
+	if (!(llcx = lookup_llc_ctx(cpuc->llc_id)))
+		goto check_llc_dsq;
+
+	if (llcx && llcx->cpumask) {
+		s32 other_cpu;
+		bpf_for(other_cpu, 0, topo_config.nr_cpus) {
+			struct bpf_cpumask *llc_cpumask;
+
+			if (other_cpu == cpu)
+				continue;
+
+			llc_cpumask = llcx->cpumask;
+			if (!llc_cpumask)
+				continue;
+
+			if (!bpf_cpumask_test_cpu(other_cpu, cast_mask(llc_cpumask)))
+				continue;
+
+			struct cpu_ctx *other_cpuc = lookup_cpu_ctx(other_cpu);
+			if (!other_cpuc)
+				continue;
+
+			// Peek at the other CPU's affn_dsq
+			p = __COMPAT_scx_bpf_dsq_peek(other_cpuc->affn_dsq);
+			if (p && bpf_cpumask_test_cpu(cpu, p->cpus_ptr) &&
+			    (p->scx.dsq_vtime < min_vtime || min_vtime == 0)) {
+				min_vtime = p->scx.dsq_vtime;
+				dsq_id = other_cpuc->affn_dsq;
+			}
+		}
+	}
+
+check_llc_dsq:
 	// LLC DSQ for vtime comparison
 	p = __COMPAT_scx_bpf_dsq_peek(cpuc->llc_dsq);
 	if (p && (p->scx.dsq_vtime < min_vtime || min_vtime == 0) &&
@@ -1741,7 +2158,17 @@ static void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 
 	// Migration eligible vtime
 	if (topo_config.nr_llcs > 1) {
-		if (p2dq_config.atq_enabled) {
+		if (p2dq_config.dhq_enabled) {
+			pid = scx_dhq_peek_strand(cpuc->mig_dhq, cpuc->dhq_strand);
+			if (pid && (p = bpf_task_from_pid((s32)pid))) {
+				if (likely(bpf_cpumask_test_cpu(cpu, p->cpus_ptr)) &&
+				    (p->scx.dsq_vtime < min_vtime || min_vtime == 0)) {
+					min_vtime = p->scx.dsq_vtime;
+					min_dhq = cpuc->mig_dhq;
+				}
+				bpf_task_release(p);
+			}
+		} else if (p2dq_config.atq_enabled) {
 			pid = scx_atq_peek(cpuc->mig_atq);
 			if ((p = bpf_task_from_pid((s32)pid))) {
 				if (likely(bpf_cpumask_test_cpu(cpu, p->cpus_ptr)) &&
@@ -1755,7 +2182,6 @@ static void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 					 * the potential cost of having to reenqueue
 					 * popped tasks if they don't match.
 					 */
-					peeked_pid = p->pid;
 				}
 				bpf_task_release(p);
 			}
@@ -1770,17 +2196,42 @@ static void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 		}
 	}
 
-	if (dsq_id != 0)
-		trace("DISPATCH cpu[%d] min_vtime %llu dsq_id %llu atq %llu",
-		      cpu, min_vtime, dsq_id, min_atq);
+	bpf_rcu_read_unlock();
 
-	// First try the DSQ with the lowest vtime for fairness.
-	if (unlikely(min_atq)) {
+	if (dsq_id != 0)
+		trace("DISPATCH cpu[%d] min_vtime %llu dsq_id %llu atq %llu dhq %llu",
+		      cpu, min_vtime, dsq_id, min_atq, min_dhq);
+
+	// First try the DHQ/ATQ with the lowest vtime for fairness.
+	if (unlikely(min_dhq)) {
+		trace("DHQ dispatching %llu with min vtime %llu", min_dhq, min_vtime);
+		pid = scx_dhq_pop_strand(min_dhq, cpuc->dhq_strand);
+		if (likely(pid && (p = bpf_task_from_pid((s32)pid)))) {
+			if (unlikely(!(taskc = lookup_task_ctx(p)))) {
+				bpf_task_release(p);
+				scx_bpf_error("DHQ failed to get task ctx");
+				return;
+			}
+
+			/* Check if task can still run on current CPU */
+
+			/* Insert to LLC DSQ for atomic affinity handling */
+			scx_bpf_dsq_insert_vtime(p,
+						 cpuc->llc_dsq,
+						 taskc->slice_ns,
+						 p->scx.dsq_vtime,
+						 taskc->enq_flags);
+			bpf_task_release(p);
+
+			/* Try to dispatch - move_to_local handles affinity atomically */
+			scx_bpf_dsq_move_to_local(cpuc->llc_dsq);
+			return;
+		}
+	} else if (unlikely(min_atq)) {
 		trace("ATQ dispatching %llu with min vtime %llu", min_atq, min_vtime);
 		pid = scx_atq_pop(min_atq);
 		if (likely((p = bpf_task_from_pid((s32)pid)))) {
 			/*
-			 * Need to ensure the peeked_pid is the pid popped off
 			 * the ATQ. Otherwise there may be priority inversions.
 			 * This probably needs to be done for the DSQs as well.
 			 */
@@ -1789,28 +2240,19 @@ static void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 				scx_bpf_error("failed to get task ctx");
 				return;
 			}
-			if (p->pid == peeked_pid) {
-				scx_bpf_dsq_insert(p,
-						   SCX_DSQ_LOCAL,
-						   taskc->slice_ns,
-						   taskc->enq_flags);
-				bpf_task_release(p);
-				return;
-			} else {
-				/*
-				 * The task that was popped was already
-				 * consumed. The next task that was popped
-				 * might have a higher vtime so reenqueue it
-				 * back to the LLC DSQ.
-				 */
-				scx_bpf_dsq_insert_vtime(p,
-						   cpuc->llc_dsq,
-						   taskc->slice_ns,
-						   p->scx.dsq_vtime,
-						   taskc->enq_flags);
-				bpf_task_release(p);
-				stat_inc(P2DQ_STAT_ATQ_REENQ);
-			}
+
+
+			/* Insert to LLC DSQ for atomic affinity handling */
+			scx_bpf_dsq_insert_vtime(p,
+						 cpuc->llc_dsq,
+						 taskc->slice_ns,
+						 p->scx.dsq_vtime,
+						 taskc->enq_flags);
+			bpf_task_release(p);
+
+			/* Try to dispatch - move_to_local handles affinity atomically */
+			scx_bpf_dsq_move_to_local(cpuc->llc_dsq);
+			return;
 		}
 	} else {
 		if (likely(valid_dsq(dsq_id) && scx_bpf_dsq_move_to_local(dsq_id)))
@@ -1846,7 +2288,29 @@ static void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 			return;
 	}
 
-	if (unlikely(p2dq_config.atq_enabled)) {
+	if (unlikely(p2dq_config.dhq_enabled)) {
+		pid = scx_dhq_pop_strand(cpuc->mig_dhq, cpuc->dhq_strand);
+		if (likely(pid && (p = bpf_task_from_pid((s32)pid)))) {
+			if (unlikely(!(taskc = lookup_task_ctx(p)))) {
+				bpf_task_release(p);
+				scx_bpf_error("DHQ failed to get task ctx");
+				return;
+			}
+
+			/* Check if task can still run on current CPU */
+
+			/* Insert to LLC DSQ for atomic affinity handling */
+			scx_bpf_dsq_insert_vtime(p,
+						 cpuc->llc_dsq,
+						 taskc->slice_ns,
+						 p->scx.dsq_vtime,
+						 taskc->enq_flags);
+			bpf_task_release(p);
+
+			/* Try to dispatch - move_to_local handles affinity atomically */
+			scx_bpf_dsq_move_to_local(cpuc->llc_dsq);
+		}
+	} else if (unlikely(p2dq_config.atq_enabled)) {
 		pid = scx_atq_pop(cpuc->mig_atq);
 		if (likely((p = bpf_task_from_pid((s32)pid)))) {
 			if (unlikely(!(taskc = lookup_task_ctx(p)))) {
@@ -1854,11 +2318,19 @@ static void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 				scx_bpf_error("failed to get task ctx");
 				return;
 			}
-			scx_bpf_dsq_insert(p,
-					   SCX_DSQ_LOCAL,
-					   taskc->slice_ns,
-					   taskc->enq_flags);
+
+			/* Check if task can still run on current CPU */
+
+			/* Insert to LLC DSQ for atomic affinity handling */
+			scx_bpf_dsq_insert_vtime(p,
+						 cpuc->llc_dsq,
+						 taskc->slice_ns,
+						 p->scx.dsq_vtime,
+						 taskc->enq_flags);
 			bpf_task_release(p);
+
+			/* Try to dispatch - move_to_local handles affinity atomically */
+			scx_bpf_dsq_move_to_local(cpuc->llc_dsq);
 			return;
 		}
 	} else {
@@ -1927,11 +2399,6 @@ void BPF_STRUCT_OPS(p2dq_set_cpumask, struct task_struct *p,
 	}
 }
 
-void BPF_STRUCT_OPS(p2dq_cpu_release, s32 cpu, struct scx_cpu_release_args *args)
-{
-	scx_bpf_reenqueue_local();
-}
-
 void BPF_STRUCT_OPS(p2dq_update_idle, s32 cpu, bool idle)
 {
 	const struct cpumask *idle_cpumask;
@@ -2040,9 +2507,10 @@ static s32 p2dq_init_task_impl(struct task_struct *p, struct scx_init_task_args
 	slice_ns = scale_by_task_weight(p,
 					dsq_time_slice(p2dq_config.init_dsq_index));
 
-	taskc->enq_flags = 0;
 	taskc->llc_id = cpuc->llc_id;
 	taskc->node_id = cpuc->node_id;
+	taskc->pid = p->pid;
+
 	// Adjust starting index based on niceness
 	if (p->scx.weight == 100) {
 		taskc->dsq_index = p2dq_config.init_dsq_index;
@@ -2053,6 +2521,8 @@ static s32 p2dq_init_task_impl(struct task_struct *p, struct scx_init_task_args
 	}
 	taskc->last_dsq_index = taskc->dsq_index;
 	taskc->slice_ns = slice_ns;
+	taskc->enq_flags = 0;
+
 	if (p->cpus_ptr == &p->cpus_mask &&
 	    p->nr_cpus_allowed == topo_config.nr_cpus)
 		task_ctx_set_flag(taskc, TASK_CTX_F_ALL_CPUS);
@@ -2106,6 +2576,11 @@ static int init_llc(u32 llc_index)
 		return ret;
 	}
 
+	ret = llc_create_dhqs(llcx);
+	if (ret) {
+		return ret;
+	}
+
 	llcx->dsq = llcx->id | MAX_LLCS;
 	ret = scx_bpf_create_dsq(llcx->dsq, llcx->node_id);
 	if (ret) {
@@ -2236,6 +2711,8 @@ static s32 init_cpu(int cpu)
 	llcx->node_id = cpu_node_ids[cpu];
 	nodec->id = cpu_node_ids[cpu];
 	cpuc->mig_atq = llcx->mig_atq;
+	cpuc->mig_dhq = llcx->mig_dhq;
+	cpuc->dhq_strand = llcx->dhq_strand;
 
 	if (cpu_ctx_test_flag(cpuc, CPU_CTX_F_IS_BIG)) {
 		trace("CPU[%d] is big", cpu);
@@ -2299,12 +2776,17 @@ static bool load_balance_timer(void)
 			return false;
 		}
 
-		load_sum += llcx->load;
-		interactive_sum += llcx->intr_load;
+		/* Use PELT metrics if enabled, otherwise use simple counters */
+		u64 llc_load = p2dq_config.pelt_enabled ? llcx->util_avg : llcx->load;
+		u64 lb_llc_load = p2dq_config.pelt_enabled ? lb_llcx->util_avg : lb_llcx->load;
+		u64 llc_intr_load = p2dq_config.pelt_enabled ? llcx->intr_util_avg : llcx->intr_load;
+
+		load_sum += llc_load;
+		interactive_sum += llc_intr_load;
 
 		s64 load_imbalance = 0;
-		if(llcx->load > lb_llcx->load)
-			load_imbalance = (100 * (llcx->load - lb_llcx->load)) / llcx->load;
+		if(llc_load > lb_llc_load)
+			load_imbalance = (100 * (llc_load - lb_llc_load)) / llc_load;
 
 		u32 lb_slack = (lb_config.slack_factor > 0 ?
 				lb_config.slack_factor : LOAD_BALANCE_SLACK);
@@ -2315,7 +2797,7 @@ static bool load_balance_timer(void)
 			llcx->lb_llc_id = MAX_LLCS;
 
 		dbg("LB llcx[%u] %llu lb_llcx[%u] %llu imbalance %lli",
-		    llc_id, llcx->load, lb_llc_id, lb_llcx->load, load_imbalance);
+		    llc_id, llc_load, lb_llc_id, lb_llc_load, load_imbalance);
 	}
 
 	dbg("LB Total load %llu, Total interactive %llu",
@@ -2356,17 +2838,37 @@ reset_load:
 		if (!(llcx = lookup_llc_ctx(llc_id)))
 			return false;
 
-		llcx->load = 0;
-		llcx->intr_load = 0;
-		llcx->affn_load = 0;
+		/*
+		 * With PELT enabled, metrics decay automatically via exponential
+		 * weighting. We only reset simple counters for legacy mode.
+		 */
+		if (!p2dq_config.pelt_enabled) {
+			llcx->load = 0;
+			llcx->intr_load = 0;
+			llcx->affn_load = 0;
+		}
+
 		llcx->last_period_ns = scx_bpf_now();
-		bpf_for(j, 0, p2dq_config.nr_dsqs_per_llc) {
-			llcx->dsq_load[j] = 0;
+
+		if (!p2dq_config.pelt_enabled) {
+			bpf_for(j, 0, p2dq_config.nr_dsqs_per_llc) {
+				llcx->dsq_load[j] = 0;
+				if (llc_id == 0 && timeline_config.autoslice) {
+					if (j > 0 && dsq_time_slices[j] < dsq_time_slices[j-1]) {
+						dsq_time_slices[j] = dsq_time_slices[j-1] << p2dq_config.dsq_shift;
+					}
+					dbg("LB autoslice interactive slice %llu", dsq_time_slices[j]);
+				}
+			}
+		} else {
+			/* Even with PELT, still validate autoslice timings */
 			if (llc_id == 0 && timeline_config.autoslice) {
-				if (j > 0 && dsq_time_slices[j] < dsq_time_slices[j-1]) {
-					dsq_time_slices[j] = dsq_time_slices[j-1] << p2dq_config.dsq_shift;
+				bpf_for(j, 1, p2dq_config.nr_dsqs_per_llc) {
+					if (dsq_time_slices[j] < dsq_time_slices[j-1]) {
+						dsq_time_slices[j] = dsq_time_slices[j-1] << p2dq_config.dsq_shift;
+					}
+					dbg("LB autoslice interactive slice %llu", dsq_time_slices[j]);
 				}
-				dbg("LB autoslice interactive slice %llu", dsq_time_slices[j]);
 			}
 		}
 	}
@@ -2509,6 +3011,7 @@ static s32 p2dq_init_impl()
 
 		cpuc->llc_dsq = llcx->dsq;
 		cpuc->mig_atq = llcx->mig_atq;
+		cpuc->mig_dhq = llcx->mig_dhq;
 
 		if (p2dq_config.llc_shards > 1 && llcx->nr_shards > 1) {
 			int shard_id = cpuc->core_id % llcx->nr_shards;
@@ -2564,10 +3067,23 @@ void BPF_STRUCT_OPS(p2dq_running, struct task_struct *p)
 void BPF_STRUCT_OPS(p2dq_enqueue, struct task_struct *p __arg_trusted, u64 enq_flags)
 {
 	struct enqueue_promise pro;
+
 	async_p2dq_enqueue(&pro, p, enq_flags);
 	complete_p2dq_enqueue(&pro, p);
 }
 
+void BPF_STRUCT_OPS(p2dq_dequeue, struct task_struct *p __arg_trusted, u64 deq_flags)
+{
+	task_ctx *taskc = lookup_task_ctx(p);
+	int ret;
+
+	ret = scx_atq_cancel(&taskc->common);
+	if (ret)
+		scx_bpf_error("scx_atq_cancel returned %d", ret);
+
+	return;
+}
+
 void BPF_STRUCT_OPS(p2dq_dispatch, s32 cpu, struct task_struct *prev)
 {
 	return p2dq_dispatch_impl(cpu, prev);
@@ -2586,8 +3102,8 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
 
 SCX_OPS_DEFINE(p2dq,
 	       .select_cpu		= (void *)p2dq_select_cpu,
-	       .cpu_release		= (void *)p2dq_cpu_release,
 	       .enqueue			= (void *)p2dq_enqueue,
+	       .dequeue			= (void *)p2dq_dequeue,
 	       .dispatch		= (void *)p2dq_dispatch,
 	       .running			= (void *)p2dq_running,
 	       .stopping		= (void *)p2dq_stopping,
