@@ -11,6 +11,21 @@ use serde::{Deserialize, Serialize};
 use std::collections::HashMap;
 use std::sync::Arc;
 
+/// Aggregation mode for task state analysis
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Serialize, Deserialize)]
+pub enum AggregationMode {
+    /// Show per-thread statistics (TID level)
+    PerThread,
+    /// Aggregate by process (TGID level) - combines all threads in a process
+    PerProcess,
+}
+
+impl Default for AggregationMode {
+    fn default() -> Self {
+        Self::PerThread
+    }
+}
+
 /// Analyzes task state transitions and time spent in each state
 pub struct TaskStateAnalyzer {
     trace: Arc<PerfettoTrace>,
@@ -21,22 +36,40 @@ impl TaskStateAnalyzer {
         Self { trace }
     }
 
-    /// Analyze task states for all or specific process
-    pub fn analyze_task_states(&self, pid_filter: Option<i32>) -> Vec<TaskStateStats> {
+    /// Analyze task states for all or specific process/thread
+    pub fn analyze_task_states(
+        &self,
+        pid_filter: Option<i32>,
+        aggregation_mode: AggregationMode,
+    ) -> Vec<TaskStateStats> {
         let mut task_trackers: HashMap<i32, TaskStateTracker> = HashMap::new();
 
-        // Process all sched_switch and wakeup events
-        for cpu in 0..self.trace.num_cpus() {
-            let events = self.trace.get_events_by_cpu(cpu as u32);
+        // Get trace duration for final stats calculation
+        let (start_ts, end_ts) = self.trace.time_range();
+        let trace_duration_ns = end_ts.saturating_sub(start_ts);
 
-            for event_with_idx in events {
-                if let Some(ts) = event_with_idx.event.timestamp {
-                    match &event_with_idx.event.event {
-                        Some(ftrace_event::Event::SchedSwitch(switch)) => {
-                            // Handle prev task being scheduled off
-                            if let (Some(prev_pid), Some(prev_state)) =
-                                (switch.prev_pid, switch.prev_state)
-                            {
+        // Collect all events from all CPUs and sort by timestamp
+        // This is critical - processing events per-CPU causes time to jump backwards
+        // for threads that migrate between CPUs
+        let mut all_events = Vec::new();
+        for cpu in 0..self.trace.num_cpus() {
+            let cpu_events = self.trace.get_events_by_cpu(cpu as u32);
+            all_events.extend(cpu_events.iter());
+        }
+
+        // Sort all events by timestamp
+        all_events.sort_by_key(|e| e.event.timestamp.unwrap_or(0));
+
+        // Process events in chronological order
+        for event_with_idx in all_events {
+            if let Some(ts) = event_with_idx.event.timestamp {
+                match &event_with_idx.event.event {
+                    Some(ftrace_event::Event::SchedSwitch(switch)) => {
+                        // Handle prev task being scheduled off
+                        if let (Some(prev_pid), Some(prev_state)) =
+                            (switch.prev_pid, switch.prev_state)
+                        {
+                            if prev_pid > 0 {
                                 let tracker = task_trackers.entry(prev_pid).or_insert_with(|| {
                                     TaskStateTracker::new(
                                         prev_pid,
@@ -44,12 +77,6 @@ impl TaskStateAnalyzer {
                                     )
                                 });
 
-                                if let Some(running_since) = tracker.running_since {
-                                    let runtime = ts.saturating_sub(running_since);
-                                    tracker.running_time_ns += runtime;
-                                }
-                                tracker.running_since = None;
-
                                 // Classify task state when scheduled off
                                 let new_state = classify_task_state(prev_state);
                                 tracker.enter_state(new_state, ts);
@@ -61,76 +88,209 @@ impl TaskStateAnalyzer {
                                     tracker.voluntary_switches += 1;
                                 }
                             }
+                        }
 
-                            // Handle next task being scheduled on
-                            if let Some(next_pid) = switch.next_pid {
-                                if next_pid > 0 {
-                                    let tracker =
-                                        task_trackers.entry(next_pid).or_insert_with(|| {
-                                            TaskStateTracker::new(
-                                                next_pid,
-                                                switch.next_comm.clone().unwrap_or_default(),
-                                            )
-                                        });
+                        // Handle next task being scheduled on
+                        if let Some(next_pid) = switch.next_pid {
+                            if next_pid > 0 {
+                                let tracker = task_trackers.entry(next_pid).or_insert_with(|| {
+                                    TaskStateTracker::new(
+                                        next_pid,
+                                        switch.next_comm.clone().unwrap_or_default(),
+                                    )
+                                });
 
-                                    // Calculate scheduler latency if task was runnable
-                                    if let Some(runnable_since) = tracker.runnable_since {
-                                        let latency = ts.saturating_sub(runnable_since);
-                                        tracker.runnable_time_ns += latency;
-                                        tracker.scheduler_latencies.push(latency);
-                                    }
-
-                                    tracker.running_since = Some(ts);
-                                    tracker.runnable_since = None;
-                                    tracker.exit_state(TaskState::Runnable, ts);
+                                // Calculate scheduler latency if we have a wakeup timestamp
+                                if let Some(wakeup_ts) = tracker.last_wakeup_ts {
+                                    let latency = ts.saturating_sub(wakeup_ts);
+                                    tracker.scheduler_latencies.push(latency);
                                 }
+
+                                // Enter running state
+                                tracker.enter_state(TaskState::Running, ts);
+                                tracker.last_wakeup_ts = None;
                             }
                         }
-                        Some(ftrace_event::Event::SchedWakeup(wakeup)) => {
-                            if let Some(pid) = wakeup.pid {
-                                if pid > 0 {
-                                    let tracker = task_trackers.entry(pid).or_insert_with(|| {
-                                        TaskStateTracker::new(
-                                            pid,
-                                            wakeup.comm.clone().unwrap_or_default(),
-                                        )
-                                    });
-
-                                    tracker.exit_current_state(ts);
-                                    tracker.runnable_since = Some(ts);
-                                }
-                            }
-                        }
-                        Some(ftrace_event::Event::SchedWaking(waking)) => {
-                            if let Some(pid) = waking.pid {
-                                if pid > 0 {
-                                    let tracker = task_trackers.entry(pid).or_insert_with(|| {
-                                        TaskStateTracker::new(
-                                            pid,
-                                            waking.comm.clone().unwrap_or_default(),
-                                        )
-                                    });
-
-                                    tracker.exit_current_state(ts);
-                                    tracker.runnable_since = Some(ts);
-                                }
-                            }
-                        }
-                        _ => {}
                     }
+                    Some(ftrace_event::Event::SchedWakeup(wakeup)) => {
+                        if let Some(pid) = wakeup.pid {
+                            if pid > 0 {
+                                let tracker = task_trackers.entry(pid).or_insert_with(|| {
+                                    TaskStateTracker::new(
+                                        pid,
+                                        wakeup.comm.clone().unwrap_or_default(),
+                                    )
+                                });
+
+                                // Enter runnable state
+                                tracker.enter_state(TaskState::Runnable, ts);
+                                tracker.last_wakeup_ts = Some(ts);
+                            }
+                        }
+                    }
+                    Some(ftrace_event::Event::SchedWaking(waking)) => {
+                        if let Some(pid) = waking.pid {
+                            if pid > 0 {
+                                let tracker = task_trackers.entry(pid).or_insert_with(|| {
+                                    TaskStateTracker::new(
+                                        pid,
+                                        waking.comm.clone().unwrap_or_default(),
+                                    )
+                                });
+
+                                // Enter runnable state
+                                tracker.enter_state(TaskState::Runnable, ts);
+                                tracker.last_wakeup_ts = Some(ts);
+                            }
+                        }
+                    }
+                    _ => {}
                 }
             }
         }
 
-        // Convert to stats
-        let mut stats: Vec<TaskStateStats> = task_trackers
+        // Convert to stats with TGID information
+        let thread_stats: Vec<TaskStateStats> = task_trackers
             .into_iter()
             .filter(|(p, _)| pid_filter.is_none_or(|filter| &filter == p))
-            .map(|(_, tracker)| tracker.into_stats())
+            .map(|(tid, tracker)| {
+                let tgid = self.trace.get_tgid_for_tid(tid);
+                tracker.into_stats(end_ts, trace_duration_ns, tgid)
+            })
             .collect();
 
-        stats.sort_by(|a, b| b.total_time_ns.cmp(&a.total_time_ns));
-        stats
+        // If aggregating by process, combine all threads for each TGID
+        let stats = match aggregation_mode {
+            AggregationMode::PerThread => thread_stats,
+            AggregationMode::PerProcess => Self::aggregate_by_tgid(thread_stats, trace_duration_ns),
+        };
+
+        let mut sorted_stats = stats;
+        sorted_stats.sort_by(|a, b| b.total_time_ns.cmp(&a.total_time_ns));
+        sorted_stats
+    }
+
+    /// Aggregate thread stats by TGID (process ID)
+    fn aggregate_by_tgid(
+        thread_stats: Vec<TaskStateStats>,
+        trace_duration_ns: u64,
+    ) -> Vec<TaskStateStats> {
+        let mut process_map: HashMap<i32, Vec<TaskStateStats>> = HashMap::new();
+
+        // Group threads by TGID
+        for stat in thread_stats {
+            let tgid = stat.tgid.unwrap_or(stat.pid);
+            process_map.entry(tgid).or_default().push(stat);
+        }
+
+        // Aggregate stats for each process
+        process_map
+            .into_iter()
+            .map(|(tgid, threads)| {
+                let thread_count = threads.len();
+                let comm = threads[0].comm.clone();
+
+                // Sum up all times across threads
+                let mut running_time_ns = 0u64;
+                let mut runnable_time_ns = 0u64;
+                let mut sleeping_time_ns = 0u64;
+                let mut blocked_time_ns = 0u64;
+                let mut voluntary_switches = 0usize;
+                let mut involuntary_switches = 0usize;
+                // Note: We can't reconstruct individual latencies from aggregated stats
+                // So we'll use weighted average latencies from the threads instead
+
+                for thread in &threads {
+                    running_time_ns += thread.running_time_ns;
+                    runnable_time_ns += thread.runnable_time_ns;
+                    sleeping_time_ns += thread.sleeping_time_ns;
+                    blocked_time_ns += thread.blocked_time_ns;
+                    voluntary_switches += thread.voluntary_switches;
+                    involuntary_switches += thread.involuntary_switches;
+                    // Note: We can't reconstruct individual latencies from aggregated stats
+                    // So we'll use average latencies from the threads
+                }
+
+                let total_time_ns =
+                    running_time_ns + runnable_time_ns + sleeping_time_ns + blocked_time_ns;
+                let total_time_clamped = total_time_ns.min(trace_duration_ns * thread_count as u64);
+
+                // Calculate aggregate scheduler latency (weighted average)
+                let mut total_latency_weight = 0u64;
+                let mut weighted_avg_latency = 0f64;
+                let mut weighted_p50_latency = 0f64;
+                let mut weighted_p95_latency = 0f64;
+                let mut weighted_p99_latency = 0f64;
+
+                for thread in &threads {
+                    let weight = thread.total_time_ns;
+                    total_latency_weight += weight;
+                    weighted_avg_latency += thread.avg_scheduler_latency_ns as f64 * weight as f64;
+                    weighted_p50_latency += thread.p50_scheduler_latency_ns as f64 * weight as f64;
+                    weighted_p95_latency += thread.p95_scheduler_latency_ns as f64 * weight as f64;
+                    weighted_p99_latency += thread.p99_scheduler_latency_ns as f64 * weight as f64;
+                }
+
+                let avg_latency = if total_latency_weight > 0 {
+                    (weighted_avg_latency / total_latency_weight as f64) as u64
+                } else {
+                    0
+                };
+                let p50_latency = if total_latency_weight > 0 {
+                    (weighted_p50_latency / total_latency_weight as f64) as u64
+                } else {
+                    0
+                };
+                let p95_latency = if total_latency_weight > 0 {
+                    (weighted_p95_latency / total_latency_weight as f64) as u64
+                } else {
+                    0
+                };
+                let p99_latency = if total_latency_weight > 0 {
+                    (weighted_p99_latency / total_latency_weight as f64) as u64
+                } else {
+                    0
+                };
+
+                TaskStateStats {
+                    pid: tgid,
+                    tgid: Some(tgid),
+                    thread_count,
+                    comm,
+                    running_time_ns,
+                    runnable_time_ns,
+                    sleeping_time_ns,
+                    blocked_time_ns,
+                    total_time_ns: total_time_clamped,
+                    running_percent: if total_time_clamped > 0 {
+                        (running_time_ns as f64 / total_time_clamped as f64) * 100.0
+                    } else {
+                        0.0
+                    },
+                    runnable_percent: if total_time_clamped > 0 {
+                        (runnable_time_ns as f64 / total_time_clamped as f64) * 100.0
+                    } else {
+                        0.0
+                    },
+                    sleeping_percent: if total_time_clamped > 0 {
+                        (sleeping_time_ns as f64 / total_time_clamped as f64) * 100.0
+                    } else {
+                        0.0
+                    },
+                    blocked_percent: if total_time_clamped > 0 {
+                        (blocked_time_ns as f64 / total_time_clamped as f64) * 100.0
+                    } else {
+                        0.0
+                    },
+                    voluntary_switches,
+                    involuntary_switches,
+                    avg_scheduler_latency_ns: avg_latency,
+                    p50_scheduler_latency_ns: p50_latency,
+                    p95_scheduler_latency_ns: p95_latency,
+                    p99_scheduler_latency_ns: p99_latency,
+                }
+            })
+            .collect()
     }
 }
 
@@ -157,10 +317,10 @@ fn classify_task_state(state: i64) -> TaskState {
 struct TaskStateTracker {
     pid: i32,
     comm: String,
-    running_since: Option<u64>,
-    runnable_since: Option<u64>,
-    sleeping_since: Option<u64>,
-    blocked_since: Option<u64>,
+    // Current state - only one state can be active at a time
+    current_state: Option<TaskState>,
+    current_state_since: Option<u64>,
+    // Accumulated times
     running_time_ns: u64,
     runnable_time_ns: u64,
     sleeping_time_ns: u64,
@@ -168,6 +328,8 @@ struct TaskStateTracker {
     scheduler_latencies: Vec<u64>,
     voluntary_switches: usize,
     involuntary_switches: usize,
+    // Track last wakeup time for latency calculation
+    last_wakeup_ts: Option<u64>,
 }
 
 impl TaskStateTracker {
@@ -175,10 +337,8 @@ impl TaskStateTracker {
         Self {
             pid,
             comm,
-            running_since: None,
-            runnable_since: None,
-            sleeping_since: None,
-            blocked_since: None,
+            current_state: None,
+            current_state_since: None,
             running_time_ns: 0,
             runnable_time_ns: 0,
             sleeping_time_ns: 0,
@@ -186,68 +346,54 @@ impl TaskStateTracker {
             scheduler_latencies: Vec::new(),
             voluntary_switches: 0,
             involuntary_switches: 0,
+            last_wakeup_ts: None,
         }
     }
 
-    fn enter_state(&mut self, state: TaskState, ts: u64) {
-        match state {
-            TaskState::Runnable => self.runnable_since = Some(ts),
-            TaskState::Sleeping => self.sleeping_since = Some(ts),
-            TaskState::Blocked => self.blocked_since = Some(ts),
-            TaskState::Running => self.running_since = Some(ts),
-        }
-    }
-
-    fn exit_state(&mut self, state: TaskState, ts: u64) {
-        match state {
-            TaskState::Runnable => {
-                if let Some(since) = self.runnable_since {
-                    self.runnable_time_ns += ts.saturating_sub(since);
-                    self.runnable_since = None;
-                }
-            }
-            TaskState::Sleeping => {
-                if let Some(since) = self.sleeping_since {
-                    self.sleeping_time_ns += ts.saturating_sub(since);
-                    self.sleeping_since = None;
-                }
-            }
-            TaskState::Blocked => {
-                if let Some(since) = self.blocked_since {
-                    self.blocked_time_ns += ts.saturating_sub(since);
-                    self.blocked_since = None;
-                }
-            }
-            TaskState::Running => {
-                if let Some(since) = self.running_since {
-                    self.running_time_ns += ts.saturating_sub(since);
-                    self.running_since = None;
-                }
-            }
-        }
-    }
-
+    /// Exit current state and accumulate time
     fn exit_current_state(&mut self, ts: u64) {
-        if self.running_since.is_some() {
-            self.exit_state(TaskState::Running, ts);
-        }
-        if self.runnable_since.is_some() {
-            self.exit_state(TaskState::Runnable, ts);
-        }
-        if self.sleeping_since.is_some() {
-            self.exit_state(TaskState::Sleeping, ts);
-        }
-        if self.blocked_since.is_some() {
-            self.exit_state(TaskState::Blocked, ts);
+        if let (Some(state), Some(since)) = (self.current_state, self.current_state_since) {
+            let duration = ts.saturating_sub(since);
+            match state {
+                TaskState::Running => self.running_time_ns += duration,
+                TaskState::Runnable => self.runnable_time_ns += duration,
+                TaskState::Sleeping => self.sleeping_time_ns += duration,
+                TaskState::Blocked => self.blocked_time_ns += duration,
+            }
         }
+        self.current_state = None;
+        self.current_state_since = None;
     }
 
-    fn into_stats(self) -> TaskStateStats {
+    /// Enter a new state (exits current state first)
+    fn enter_state(&mut self, state: TaskState, ts: u64) {
+        // Exit current state before entering new one
+        self.exit_current_state(ts);
+
+        self.current_state = Some(state);
+        self.current_state_since = Some(ts);
+    }
+
+    fn into_stats(
+        mut self,
+        trace_end_ts: u64,
+        trace_duration_ns: u64,
+        tgid: Option<i32>,
+    ) -> TaskStateStats {
+        // If still in a state at end of trace, close it out
+        if self.current_state.is_some() {
+            self.exit_current_state(trace_end_ts);
+        }
+
         let total_time = self.running_time_ns
             + self.runnable_time_ns
             + self.sleeping_time_ns
             + self.blocked_time_ns;
 
+        // Validation: total time should not exceed trace duration
+        // If it does, it indicates a bug in our accounting
+        let total_time_clamped = total_time.min(trace_duration_ns);
+
         let scheduler_latency_percentiles = if !self.scheduler_latencies.is_empty() {
             PerfettoTrace::calculate_percentiles(&self.scheduler_latencies)
         } else {
@@ -256,29 +402,31 @@ impl TaskStateTracker {
 
         TaskStateStats {
             pid: self.pid,
+            tgid,
+            thread_count: 1, // Single thread
             comm: self.comm,
             running_time_ns: self.running_time_ns,
             runnable_time_ns: self.runnable_time_ns,
             sleeping_time_ns: self.sleeping_time_ns,
             blocked_time_ns: self.blocked_time_ns,
-            total_time_ns: total_time,
-            running_percent: if total_time > 0 {
-                (self.running_time_ns as f64 / total_time as f64) * 100.0
+            total_time_ns: total_time_clamped,
+            running_percent: if total_time_clamped > 0 {
+                (self.running_time_ns as f64 / total_time_clamped as f64) * 100.0
             } else {
                 0.0
             },
-            runnable_percent: if total_time > 0 {
-                (self.runnable_time_ns as f64 / total_time as f64) * 100.0
+            runnable_percent: if total_time_clamped > 0 {
+                (self.runnable_time_ns as f64 / total_time_clamped as f64) * 100.0
             } else {
                 0.0
             },
-            sleeping_percent: if total_time > 0 {
-                (self.sleeping_time_ns as f64 / total_time as f64) * 100.0
+            sleeping_percent: if total_time_clamped > 0 {
+                (self.sleeping_time_ns as f64 / total_time_clamped as f64) * 100.0
             } else {
                 0.0
             },
-            blocked_percent: if total_time > 0 {
-                (self.blocked_time_ns as f64 / total_time as f64) * 100.0
+            blocked_percent: if total_time_clamped > 0 {
+                (self.blocked_time_ns as f64 / total_time_clamped as f64) * 100.0
             } else {
                 0.0
             },
@@ -294,7 +442,12 @@ impl TaskStateTracker {
 
 #[derive(Debug, Clone, Serialize, Deserialize)]
 pub struct TaskStateStats {
+    /// TID (thread ID) for per-thread stats, or TGID for per-process stats
     pub pid: i32,
+    /// TGID (thread group ID / process ID) - same as pid for per-thread, different for per-process
+    pub tgid: Option<i32>,
+    /// Number of threads aggregated (1 for per-thread, >1 for per-process)
+    pub thread_count: usize,
     pub comm: String,
     pub running_time_ns: u64,
     pub runnable_time_ns: u64,
